from __future__ import print_function

from .sage_types import Graphics, floor

def get_hasse(P):
    G = Graphics()
    FL = P.face_lattice()
    HD = FL.hasse_diagram()
    pos_dict, size_dict = get_vert_dicts(HD,P)
    HD.set_pos(pos_dict)
    half_dim = floor(P.dim()/2)
    w=len(P.faces(half_dim))
#    HD.set_latex_options(graphic_size=((P.dim()+3)*(w/4.),(P.dim()+3)*(w/6.)), vertex_size=1, edge_thickness=0.05, tkz_style="Shade",vertex_color="red")
    HD = HD.to_undirected()
    verts_by_dim = dict([(i,[v for v in HD.vertices() if v.dim()==i]) for i in range(-1,P.dim()+1)])
    min_dimension = min([i for i in verts_by_dim if len(verts_by_dim[i])!=0])
    for i in range(min_dimension,P.dim()+1):
        subgraph = HD.subgraph(verts_by_dim[i])
        subgraph.set_pos(dict([(v,pos_dict[v]) for v in verts_by_dim[i]]))
        size = size_dict[i]
        subgraph.relabel(relabel)
        if i==P.dim():
            p = subgraph.plot(dpi=200,vertex_size=200,vertex_labels=False,vertex_color="black")
        else:
            p = subgraph.plot(dpi=200,vertex_size=200*(size**1.7))
        G+=p
    HD.relabel(relabel)
    p = HD.plot(dpi=200,vertex_size=0,edge_thickness=1,vertex_labels=False)
    G+=p
    G.SHOW_OPTIONS['figsize'] = ((P.dim()+3)*(w/6.),(P.dim()+3)*(w/9.))
    G.axes(False)
    return G


def relabel(v):
    verts = v.vertices()
    length = len(verts)
    label = '{},'*(length-1)+'{}' if length>=2 else '{}'*length
#    label = '$\\text{{'+'{},'*(length-1)+'{}'+'}}$' if length>=2 else '$'+'{}'*length+'$'
    label = label.format(*[vert.index() for vert in verts])
    return label


def get_vert_dicts(G,P):
    dim = P.dim()
    verts_by_dim = dict([(i,[v for v in G.vertices() if v.dim()==i]) for i in range(-1,dim+1)])
    pos_dict = {}
    size_dict = {}
    levels = dim + 2
    levs = [1*i for i in range(levels)]
    v_top = verts_by_dim[dim][0]
    v_null = verts_by_dim[-1][0]
    pos_dict[v_top] = (.5,1-1./(dim+3))
    pos_dict[v_null] = (.5,1./(dim+3))
    size_dict[dim] = 1
    size_dict[-1] = 1
    half_dim = floor(dim/2.)
    g = 1./(len(verts_by_dim[half_dim]))
    for i in range(0,dim):
        verts = [v for v in G.vertices() if v in verts_by_dim[i]]
        verts.sort()
        maxlen = max([len(v.vertices()) for v in verts])
        pad = 2*abs(half_dim-i)*g + 2*g
        gap = (1-pad)/(len(verts)-1)
        if gap < g:
            gap = 1./len(verts)
            pad = gap
        size_dict[i] = maxlen
        for j in range(len(verts)):
            pos_dict[verts[j]] = (pad/2.+gap*j,(1./(dim+3)*(2+i)))
    return pos_dict, size_dict



#    [[1,1/3, 1/3, 1/3, 0],[1,1, 0, 0, 0],[1,0, 1/2, 0, 0],[1,0, 1/3, -1/3, -1/3],[1,0, 0, 1/2, 0],[1,0, 0, 0, 1],[1,0, 0, 0, -1],[1,0, 0, -1/2, 0],[1,0, -1/3, 1/3, 1/3],[1,0, -1/2, 0, 0],[1,-1, 0, 0, 0],[1,-1/3, -1/3, -1/3, 0]]




from __future__ import print_function

import tkinter as tk
from tkinter import ttk


def generate_summary(tnorm_app):
    try:
        print('Generating summary page... ', end='')
        W=tnorm_app.wrapper
        B=tnorm_app.ball
        P=tnorm_app.ball.polyhedron()
        M=tnorm_app.wrapper.manifold()
        try:
            sym_group = M.symmetry_group()
        except ValueError:
            sym_group = 'Failed'
        sections = ['Manifold', 'Thurston norm ball', 'Normal surfaces']
        names = [('rank H_2(M;bdy M)', 'volume', 'num tetrahedra', 'num cusps', 'symmetry group'),('num vertices', 'vertex degrees', 'vertices qtons rep (genus,punctures)', 'num faces', 'top dim faces (<verts>)'),('allows non-admissible', 'num qtons', 'num non-trivial in H2', 'num norm minimizing')]
        qtons_inds = [v.qtons_index() for v in B.vertices()]
        values = [(M.homology().rank(), M.volume(), M.num_tetrahedra(), M.num_cusps(), sym_group), (len(P.vertices()), [sum(P.vertex_adjacency_matrix()[i]) for i in range(P.n_vertices())], [(v.genus(),v.num_boundary_comps()) for v in B.vertices()], len(P.faces(P.dim()-1)), ['<{}>'.format(' '.join([str(v.index()) for v in face.vertices()])) for face in P.faces(P.dim()-1)]), (W.allows_non_admissible(), W.qtons().size(), len([1 for i in range(W.qtons().size()) if W.over_facet(i)!=None]), len([1 for i in range(W.qtons().size()) if W.is_norm_minimizing(i)]))]
        summary_page(tnorm_app.SummaryTab, sections, names, values)
        print('Done.')
    except Exception as e: print('Error: {}'.format(e))
    tnorm_app.stop_spin()

def generate_nb_overview(tnorm_app):
    try:
        W=tnorm_app.wrapper
        B=tnorm_app.ball
        P=tnorm_app.ball.polyhedron()
        M=tnorm_app.wrapper.manifold()

        sections = ['Overview']
        names = [('dimension', 'is compact?', 'num vertices', 'num rays', 'vertex degrees', 'vertices qtons rep (genus,punctures)', 'all vertices admissible?', 'num faces', 'faces (<vertices by index>)')]
        values = [(B.dimension(), B.is_compact(), len(P.vertices()), B.num_rays(), [sum(P.vertex_adjacency_matrix()[i]) for i in range(P.n_vertices())], [(v.genus(),v.num_boundary_comps()) for v in B.vertices()], B.all_vertices_admissible(), len(P.faces(P.dim()-1)), ['<{}>'.format(' '.join([str(B.index_of_poly_vert(v)) for v in face.vertices()])) for face in P.faces(P.dim()-1)])]
        summary_page(tnorm_app.NormBallTab, sections, names, values)
    except Exception as e: print('Error: {}'.format(e))
    tnorm_app.stop_spin()

def generate_dnb_overview(tnorm_app):
    try:
        W=tnorm_app.wrapper
        B=tnorm_app.dual_ball
        P=tnorm_app.dual_ball.polyhedron()
        M=tnorm_app.wrapper.manifold()

        sections = ['Overview']
        names = [('dimension', 'is full dimensional?', 'num vertices', 'vertex degrees', 'num faces', 'faces (<vertices by index>)')]
        values = [(B.dimension(), B.is_full_dimensional(), len(P.vertices()), [sum(P.vertex_adjacency_matrix()[i]) for i in range(P.n_vertices())], len(P.faces(P.dim()-1)), ['<{}>'.format(' '.join([str(B.index_of_poly_vert(v)) for v in face.vertices()])) for face in P.faces(P.dim()-1)])]
        summary_page(tnorm_app.DualNormBallTab, sections, names, values)
    except Exception as e: print('Error: {}'.format(e))
    tnorm_app.stop_spin()

def summary_page(frame, sections, names, values):

    col1_width = max([len('{}'.format(section_names[i])) for section_names in names for i in range(len(section_names))]) + 5
    col2_width = max([len('{}'.format(section_values[i])) for section_values in values for i in range(len(section_values))]) + 5
    
    text = tk.Text(frame, wrap='none', bg="#EAEAEA", height=0, borderwidth=-2, padx=-1, pady=-1, insertwidth=0)
    xscrollbar = ttk.Scrollbar(frame, orient='horizontal')
    yscrollbar = ttk.Scrollbar(frame, orient='vertical')
    xscrollbar.config(command=text.xview)
    yscrollbar.config(command=text.yview)
    text['xscrollcommand'] = xscrollbar.set
    text['yscrollcommand'] = yscrollbar.set
    frame.columnconfigure(0, weight=1) # column with treeview
    frame.rowconfigure(0, weight=1) # row with treeview  
    yscrollbar.grid(row=0, column=1, sticky='ns')
    text.grid(row=0, column=0, sticky='nswe')
    xscrollbar.grid(row=1, column=0, sticky='we')
    column1 = tk.Text(text, state='normal', wrap='none', width = 44, bg="#EAEAEA", borderwidth=-2, padx=-1, pady=-1)
    column2 = tk.Text(text, state='normal', wrap='none', width = col2_width, bg="#EAEAEA", borderwidth=-2, padx=-1, pady=-1)
    text.window_create(tk.END, window=column1)
    text.window_create(tk.END, window=column2)
    
    column1.tag_add('left', tk.INSERT)
    column2.tag_add('left', tk.INSERT)
    column1.tag_add('right', tk.INSERT)
    column1.tag_add('right', tk.INSERT)
    column1.tag_configure('left', justify='left')
    column2.tag_configure('left', justify='left')
    column1.tag_configure('right', justify='right')
    column1.tag_configure('right', justify='right')
    column1.bind("<1>", lambda event: column1.focus_set())
    column2.bind("<1>", lambda event: column2.focus_set())

    height = 1
    for i in range(len(sections)):
        column1.insert(tk.END, '\n\n   {}:{}\n'.format(sections[i], ' '*(col1_width-5-len(sections[i]))), ('left'))
        column2.insert(tk.END, '\n\n{}\n'.format(' '*(col2_width-2)), ('left'))
        height += 3
        for j in range(len(names[i])):
            height += 1
            column1.insert(tk.END, '{}:\n'.format(names[i][j]), ('right'))
            column2.insert(tk.END, ' {}\n'.format(values[i][j]), ('left'))
    column1.config(state='disable', height=height)
    column2.config(state='disable', height=height)










from tnorm.kernel import matrices
from tnorm.utilities.regina_helpers import regina_to_sage_int

from tnorm.utilities.sage_types import Integer, QQ



def get_quads(spun_surface,tet,v1,v2):
    s = spun_surface
    tet = int(tet)
    v2 = int(v2)
    return regina_to_sage_int(s.quads(tet,v2-int(1)))


def signed_bdy_maps(oriented_spun_surface, peripheral_curve_mats, oriented_quads_mat=None, spinning=False):
    pass

### this maps an oriented spun normal surface to H_1(bdy M; \ZZ)
def inward_oriented_bdy(qtons, peripheral_curve_mats, oriented_quads_mat=None, spinning=False):
    s = qtons
    T = s.triangulation()
    quads_mat = oriented_quads_mat


    neg_intx_mat = matrices.NEG_INTERSECTION_MAT
    if spinning:
        neg_intx_mat = matrices.ABSNEG_INTERSECTION_MAT
    neg_slopes = []
    if quads_mat == None:
        quads_mat = matrices.oriented_quads_mat(s)

    neg_intx_times_quad_mat = (neg_intx_mat*quads_mat.transpose()).transpose()

    for c in range(T.countCusps()):
        periph_mat = peripheral_curve_mats[c] # get the matrix that encodes meridian and longitude for cusp c.
        
        iota_lambda_neg=0
        iota_mu_neg=0

        for i in range(T.size()):
            iota_lambda_neg += Integer(periph_mat[0][i]*neg_intx_times_quad_mat[i])
            iota_mu_neg += Integer(periph_mat[1][i]*neg_intx_times_quad_mat[i])

  
        neg_slopes.append((-iota_lambda_neg,iota_mu_neg))
    return neg_slopes


def outward_oriented_bdy(qtons, peripheral_curve_mats, oriented_quads_mat=None, spinning=False):
    s = qtons
    T = s.triangulation()
    quads_mat = oriented_quads_mat


    pos_intx_mat = matrices.POS_INTERSECTION_MAT
    pos_slopes = []                   # on a surface affects boundary orientations.
    if quads_mat == None:
        quads_mat = matrices.oriented_quads_mat(s)

    pos_intx_times_quad_mat = (pos_intx_mat*quads_mat.transpose()).transpose()

    for c in range(T.countCusps()):
        periph_mat = peripheral_curve_mats[c] # get the matrix that encodes meridian and longitude for cusp c.
        
        iota_lambda_pos=0
        iota_mu_pos=0

        for i in range(T.size()):
            iota_lambda_pos += Integer(periph_mat[0][i]*pos_intx_times_quad_mat[i])
            iota_mu_pos += Integer(periph_mat[1][i]*pos_intx_times_quad_mat[i])

  
        pos_slopes.append((-iota_lambda_pos,iota_mu_pos))
    return pos_slopes



### this is almost the same as sign_bdy_matrix(), except we take the (entry-wise) absolute value
### of neg_intersection_mat(). This puts the orientation on quad edges described in Tillmann-- "Normal 
### surfaces in topologically finite 3-manifolds". The result are boundary slopes with signs indicating
### spinning direction.
def outward_spinning_bdy(qtons, peripheral_curve_mats, oriented_quads_mat):
    return outward_oriented_bdy(qtons, peripheral_curve_mats, oriented_quads_mat, True)

def inward_spinning_bdy(qtons, peripheral_curve_mats, oriented_quads_mat):
    return inward_oriented_bdy(qtons, peripheral_curve_mats, oriented_quads_mat, True)


### this function computes spinning slopes for (unoriented) spun normal surfaces. 

def unoriented_spinning_slopes(spun_surface, peripheral_curve_mats, quads_mat=None):
    s = spun_surface
    T = s.triangulation()

    intx_mat = matrices.UNORIENTED_INTERSECTION_MAT
    slopes = []                   # on a surface affects boundary orientations.
    if quads_mat == None:
        quads_mat = matrices.quads_mat(s)

    intx_times_quad_mat = (intx_mat*quads_mat.transpose()).transpose()

    for c in range(T.countCusps()):
        periph_mat = peripheral_curve_mats[c] # get the matrix that encodes meridian and longitude for cusp c.
        
        iota_lambda=0
        iota_mu=0

        for i in range(T.size()):
            iota_lambda += Integer(periph_mat[0][i]*intx_times_quad_mat[i])
            iota_mu += Integer(periph_mat[1][i]*intx_times_quad_mat[i])

  
        slopes.append((-iota_lambda,iota_mu))
    return slopes















from tnorm.utilities.regina_helpers import regina_to_sage_int




### Pachner moves performed with SnapPy can result in a peripheral basis that has
### trivial loops, which throws off Euler characteristic calculations. So we need to
### check that each curve in the peripheral basis provided by SnapPy is connected. We
### also check that each meridian, longitude pair has algebraic intersection number +/- 1,
### which is a suffient condition to guarantee that they form a basis.

def alg_intersections(TN_wrapper):
    W = TN_wrapper
    P = W._peripheral_curve_mats
    Tri = W.triangulation()

    alg_int_numbers = []

    for cusp in range(W.num_cusps()):
        m_mat = P[cusp][0]
        l_mat = P[cusp][1]
        alg_int = 0

        for t in range(Tri.size()):
            row_t = m_mat[t]
            for v in range(4):
                for f in range(4):
                    if row_t[4*v+f] > 0:
                        b = row_t[4*v+f]

                        left, right = neighbors(v,f)
                        l_arcs = -(b + abs(row_t[4*v+left]) - abs(row_t[4*v+right]))/2
                        r_arcs = -(b + abs(row_t[4*v+right]) - abs(row_t[4*v+left]))/2

                        if l_mat[t][4*v+right] > 0:
                            alg_int += -(l_mat[t][4*v+right])*l_arcs/2


                        elif l_mat[t][4*v+right] < 0:
                            alg_int += -(l_mat[t][4*v+right])*l_arcs/2


                        if l_mat[t][4*v+left] > 0:
                            alg_int += (l_mat[t][4*v+left])*r_arcs/2


                        elif l_mat[t][4*v+left] < 0:
                            alg_int += (l_mat[t][4*v+left])*r_arcs/2

        alg_int_numbers.append(alg_int)
    
    return alg_int_numbers


def periph_basis_intersections(TN_wrapper):
    alg_ints = alg_intersections(TN_wrapper)
    good_basis = [abs(alg_ints[i]) == 1 for i in range(len(alg_ints))]
    if all(good_basis):
        return True, good_basis
    else:
        return False, 'peripheral basis for cusp {} does not have algebraic intersection +/- 1'.format(good_basis.index(False))


def periph_basis_connected(TN_wrapper):
    W = TN_wrapper
    P = W._peripheral_curve_mats
    Tri = W.triangulation()

    components = {}

    for cusp in range(W.num_cusps()):
        for curve in range(2):
            pmat = P[cusp][curve]
            
            visited = [(t,v) for t in range(Tri.size()) for v in range(4) if sum([abs(pmat[t][4*v+k]) for k in range(4)])>0]
            pos_weights = {(t,v,f):[1 for i in range(pmat[t][4*v+f])] for (t,v) in visited for f in range(4) if pmat[t][4*v+f] > 0}
            unvisited_arcs = sum([sum(value) for value in pos_weights.values()])

            t,v = visited[0]
            weights = [pmat[t][4*v+k] for k in range(4)]
            f = first_positive(weights)
            p = 1
            arc = (int(t),int(v),int(f),int(p))
            
            def iterate(component, arc, Tri):
                next_arc = arc_map(arc,pmat, Tri)
                component.append(next_arc)

                return component, next_arc


            component = [arc]
            unvisited_arcs -= 1
            pos_weights[(t,v,f)][p-1] -= 1
            component, arc = iterate(component, arc, Tri)

            while arc != component[0]:
                if unvisited_arcs == 0:
                    return False, 'peripheral curve {} on cusp {} not closed'.format(curve, cusp)
                unvisited_arcs -= 1
                pos_weights[(arc[0],arc[1],arc[2])][arc[3]-1] -= 1
                component, arc = iterate(component, arc, Tri)

            if unvisited_arcs != 0:
                return False, 'peripheral curve {} on cusp {} not connected'.format(curve, cusp)
            assert sum([sum([abs(w) for w in value]) for value in pos_weights.values()]) == 0
            components[cusp,curve] = component
    return True, components



def neighbors(v,f):
    faces = [1,2,3] if v==0 else [0,3,2] if v==1 else [0,1,3] if v==2 else [0,2,1]
    f_index = faces.index(f)
    left, right = faces[(f_index+1)%3], faces[f_index-1]
    return left, right

def first_positive(lst):
    for i in range(len(lst)):
        if lst[i] > 0:
            return i
    return None


def arc_map(arc, pmat, Tri):
    t, v, f, p = arc

    left, right = neighbors(v,f)
    etype = 1 if pmat[t][4*v + left] >= 0 else 2 if pmat[t][4*v + right] >= 0 else 3
    if etype == 1:
        ff = right
        p0 = pmat[t][4*v+left] + p
    elif etype == 2:
        ff = left
        p0 = p
    elif etype == 3:
        b = abs(pmat[t][4*v + left])
        c = abs(pmat[t][4*v + right])
        if p <= b:
            ff = left
            p0 = p
        elif p > b:
            ff = right
            p0 = p - b

    t0, v0, f0 = glued_to(t,v,ff, Tri)

    return (int(t0), int(v0), int(f0), int(p0))


def glued_to(t, v, f, Tri):
    tet = Tri.tetrahedron(t)
    t0 = tet.adjacentTetrahedron(f).index()
    v0 = tet.adjacentGluing(f)[v]
    f0 = tet.adjacentGluing(f)[f]

    return t0, v0, f0









from __future__ import print_function

import time
import tempfile
import contextlib
import shutil
import webbrowser

from tnorm.utilities.x3d_to_html import make_x3d_html
from tnorm.utilities.hasse import get_hasse
from tnorm.utilities.cached_prop import *
from tnorm.utilities.sage_types import vector, Graphics, text3d, Polyhedron, text, RR, QQ


@contextlib.contextmanager
def make_temp_directory():
    temp_dir = tempfile.mkdtemp()
    try:
        yield temp_dir
    finally:
        shutil.rmtree(temp_dir)


class NormBall(object):
    def __init__(self,vertices, polyhedron):
        self._vertices = vertices
        self._polyhedron = polyhedron
        self._dimension = polyhedron.dim()
        self._is_full_dimensional = self._polyhedron.is_full_dimensional()
        self._is_compact = self._polyhedron.is_compact()
        self._num_vertices = len(self._polyhedron.faces(0))
        self._num_faces = len(self._polyhedron.faces(self._polyhedron.dim()-1))
        self._face_list = self._polyhedron.faces(self._polyhedron.dim()-1)
        self._face_dict = {}
        self._confirmed = False

    def vertices(self):
        return self._vertices

    def polyhedron(self):
        return self._polyhedron

    def dim(self):
        return self._dimension

    def dimension(self):
        return self._dimension

    def is_full_dimensional(self):
        return self._is_full_dimensional

    def is_compact(self):
        return self._is_compact

    def num_vertices(self):
        return self._num_vertices

    def num_faces(self):
        return self._num_faces

    def face_list(self):
        return self._face_list

    def face_dict(self):
        return self._face_dict

    ### sage does not index lines, rays, and vertices independently. We would like our indices to agree with those 
    ### of the sage Polyhedron, with lines and vertices indexed separately, so instead of using sages indices 
    ### via vertex.index(), we take the index of a vertex (or line) to be the index of the vertex (or line) in the tuple returned
    ### by self.polyhedron.vertices() (or self.polyhedron.lines()). This tuple is cached by sage, so the the order of the vertices in this 
    ### tuple should never change. The below function fetches this index.
    def index_of_poly_vert(self, v, is_line=False):   ## or line
        if is_line:
            verts = self.polyhedron().lines()
        else:
            verts = self.polyhedron().vertices()
        for i in range(len(verts)):
            if v == verts[i]:
                return i

    def vertex(self, index):
        return self.vertices()[index]

    def show_schlegel_diagram(self):
        G = Graphics()
        sp = self.polyhedron().schlegel_projection()
        p = sp.plot()
        G += p
        G.show(viewer='threejs', online=True)

    @cached_property
    def get_hasse_diagram(self):
        G = get_hasse(self.polyhedron())
        return G

    def show_hasse_diagram(self):
        G = self.get_hasse_diagram
        G.show()

    def get_hasse_graph(self):
        return self.polyhedron().face_lattice().hasse_diagram()


    def _plot2d(self, dual=False):
        labels = Graphics()
        for i in range(self.num_vertices()):
            v=self.vertices()[i]
            if not dual:
                if self.dimension() ==1:
                    labels+=text('S_{},{}'.format(v.genus(),v.num_boundary_comps()), (v.coords()[0]*1.15,0), color='black', dpi=200, fontsize='large', horizontal_alignment='center')
                else:
                    labels+=text('S_{},{}'.format(v.genus(),v.num_boundary_comps()), tuple(v.coords()[j]*1.15 for j in range(len(v.coords()))), color='black', dpi=200, fontsize='large', horizontal_alignment='center')
            else:
                if self.dimension() ==1:
                    labels+=text('{}'.format(v.index()), (v.coords()[0]*1.15,0), color='black', dpi=200, fontsize='large', horizontal_alignment='center')
                else:
                    labels+=text('{}'.format(v.index()), tuple(v.coords()[j]*1.15 for j in range(len(v.coords()))), color='black', dpi=200, fontsize='large', horizontal_alignment='center')
        plt = self.polyhedron().plot(dpi=200) + labels
        return plt

    def plot(self, viewer='x3d', online=False, transparency=.4, show_labels=True, show_axes=True, highlight_edges=True, label_color=(1,1,1), edge_color=(1,1,1), label_size=1, edge_thickness=.01, vertex_size=.06, ball_color=(0,0,1), draw_schlegel_faces=True, window_size=(1000,700)):
        dual = True if isinstance(self, DualNormBall) else False
        P = self.polyhedron()
        if not P.is_compact():
            P = Polyhedron(P.vertices_list()) # we can't plot an infinite polyhedron, so just plot the finite part.
        if P.dim() in [1,2]:
            plt = self._plot2d(dual)
            plt.show()
        elif P.dim() in [3,4]:
            if viewer == 'x3d':
                with make_temp_directory() as temp_dir:
                    filename = make_x3d_html(self, transparency=transparency, online=online, directory=temp_dir, dual=dual, show_labels=show_labels, show_axes=show_axes, highlight_edges=highlight_edges, label_color=label_color, edge_color=edge_color, label_size=label_size, edge_thickness=edge_thickness, vertex_size=vertex_size, ball_color=ball_color, draw_schlegel_faces=draw_schlegel_faces,window_size=window_size)
                    webbrowser.open('file://'+filename)
                    time.sleep(5)
            else:
                labels=Graphics()
                for i in range(self.num_vertices()):
                    v=self.vertices()[i]
                    if not dual:
                        labels+=text3d('S_{},{}'.format(v.genus(),v.num_boundary_comps()),(v.coords()[0]*1.15,v.coords()[1]*1.15,v.coords()[2]*1.15),color='black',dpi=200,fontsize='large',horizontal_alignment='center')
                    else:
                        labels+=text3d('{}'.format(v.index()),(v.coords()[0]*1.15,v.coords()[1]*1.15,v.coords()[2]*1.15),color='black',dpi=200,fontsize='large',horizontal_alignment='center')
                p = P.plot(opacity=opacity, edge_thickness=2)
                G = p+labels
                G.show(viewer=viewer,online=online,axes=True)
        else:
            print('Can\'t plot polyhedron of dimension greater than 4.')

class TNormBall(NormBall):
    def __init__(self, vertices, rays, polyhedron):
        NormBall.__init__(self, vertices, polyhedron)
        self._rays = rays
        if not self.is_compact():
            self._polyhedron_mod_rays = Polyhedron(polyhedron.vertices_list())
        else:
            self._polyhedron_mod_rays = polyhedron
        self._num_rays = len(self.polyhedron().rays())+2*len(self.polyhedron().lines())
        self._all_vertices_admissible = True
        for v in self.vertices():
             if not v.is_admissible():
                self._all_vertices_admissible = False
        
        for dim in range(1,self.polyhedron().dim()):
            self._face_dict[dim] = []
            for i in range(len(self.polyhedron().faces(dim))):
                face = self.polyhedron().faces(dim)[i]
                vertices = [self.vertex(self.index_of_poly_vert(v)) for v in face.vertices()]
                rays = [self.ray(2*self.index_of_poly_vert(l, True)+j) for l in face.lines() for j in [0,1]]
                self._face_dict[dim].append(Facet(i, dim, vertices, rays))

    def polyhedron_mod_rays(self):
        return self._polyhedron_mod_rays

    def num_rays(self):
        return self._num_rays

    def all_vertices_admissible(self):
        return self._all_vertices_admissible

    def simplicial_class_of_vertex(self,vertex_index):
        return self.vertices[vertex_index].simplicial_class

    def facets(self, dim):
        if dim == 0:
            return self.vertices()
        else:
            return self._face_dict[dim]

    def facet(self, dim, index):
        if dim == 0:
            return self.vertices()[index]
        else:
            return self._face_dict[dim][index]

    def num_facets(dim):
        return len(self.facets(dim))

    def rays(self):
        return self._rays

    def ray(self, index):
        return self._rays[index]

    def project_to_norm(self):
        """ if the norm ball is not compact, project the infinite polyhedron to
            a subspace in which the Thurston norm is a true norm. While B.polyhedron_mod_rays
            is such a projection of P, it still lives in the same ambient space as P. This 
            function returns a polyhedron that is full dimensional in its ambient space (i.e.,
            the space is projected as well as the polyhedron). If B is compact, this just
            returns P (i.e., it does nothing). Also, if the dimension of P is < 4, this just returns
            B.polyhedron_mod_rays. This is becuase 
        """
        if self.polyhedron().dim() < 4:
            return self.polyhedron_mod_rays()
        else:
            P = self.polyhedron_mod_rays()
            subspace = P.change_ring(RR).ambient_space().subspace(P.vertices_list())
            basis = subspace.basis()
            projected_verts = [[vector(v).dot_product(b.normalized()) for b in basis] for v in P.vertices_list()]
            return Polyhedron(projected_verts)

class DualNormBall(NormBall):
    def __init__(self, vertices, polyhedron):
        NormBall.__init__(self, vertices, polyhedron)
        assert self.is_compact() == True  # the dual norm ball should always be compact


class Vertex(object):
    def __init__(self, index, coords):
        self._index = index
        self._coords = coords

    def coords(self):
        return self._coords

    def index(self):
        return self._index


class NBVertex(Vertex):
    def __init__(self, index, qtons_index, coords, TN_wrapper, norm_minimizer_info=None):
        Vertex.__init__(self, index, coords)

        # the qtons surface that is a multiple of this vertex.
        self._qtons_index = qtons_index

        self._TN_wrapper = TN_wrapper

        if qtons_index != None:
            self._has_qtons_rep = True
            self._norm_minimizer_bdy_comps = self._TN_wrapper.num_boundary_comps(self._qtons_index)
            self._norm_minimizer_euler = self._TN_wrapper.euler_char(self._qtons_index)
            self._norm_minimizer_bdy_slopes = self._TN_wrapper.boundary_slopes(self._qtons_index)
            self._norm_minimizer_genus = self._TN_wrapper.genus(self._qtons_index)  
            self._norm_minimizer_is_embedded = self._TN_wrapper.is_embedded(self._qtons_index)
            self._norm_minimizer_ends_embedded = self._TN_wrapper.ends_embedded(self._qtons_index)
            self._simplicial_class = self._TN_wrapper.simplicial_class(self._qtons_index)
            self._is_admissible = self._TN_wrapper.is_admissible(self._qtons_index)    
        else:
            self._has_qtons_rep = False

            # norm_minimizer_info is a tuple containing (num_boundary_comps, euler_char, bdy_slopes).
            # Since there is no spun normal surface representing the norm minimizer, this info does not
            # correspond to a surface in TN_wrapper.qtons(), but does correspond to the norm minimizer
            # for the primitive integer point lying over the vertex.
            self._norm_minimizer_bdy_comps = norm_minimizer_info[0]
            self._norm_minimizer_euler = norm_minimizer_info[1]
            self._norm_minimizer_bdy_slopes = norm_minimizer_info[2]
            self._norm_minimizer_genus = (2-self._norm_minimizer_euler-self._norm_minimizer_bdy_comps)/2
            self._simplicial_class = None
            self._is_admissible = None
            self._norm_minimizer_is_embedded = None
            self._norm_minimizer_ends_embedded = None

        self._is_ray = False
        self._factor = QQ(-1/self.euler_char()) if self.euler_char() != 0 else float('inf')

        # the index of the spun normal surface that is norm minimizing and is the primitive integer point on the ray through this vertex.
    def qtons_index(self):
        return self._qtons_index

    def simplicial_class(self):
        """return the simplicial class of the norm minimizing surface that is a multiple of this vertex 
            (i.e., it is a primitive integer point lying on the ray over this vertex)
        """
        return self._simplicial_class

    def TN_wrapper(self):
        return self._TN_wrapper

    def is_ray(self):
        return self._is_ray

    def genus(self):
        """return the genus of the norm minimizing surface that is a multiple of this vertex (i.e., it 
            is a primitive integer point lying on the ray over this vertex)
        """
        return self._norm_minimizer_genus

    def num_boundary_comps(self):
        """return the number of boundary components of the norm minimizing surface that is a multiple of 
            this vertex (i.e., it is a primitive integer point lying on the ray over this vertex)
        """
        return self._norm_minimizer_bdy_comps

    def euler_char(self):
        """return the Euler characteristic of the norm minimizing surface that is a multiple of this vertex 
            (i.e., it is a primitive integer point lying on the ray over this vertex)
        """
        return self._norm_minimizer_euler

    def boundary_slopes(self):
        """return the boundary slopes of the norm minimizing surface that is a multiple of this vertex 
            (i.e., it is a primitive integer point lying on the ray over this vertex)
        """
        return self._norm_minimizer_bdy_slopes   

    def is_embedded(self):
        return self._norm_minimizer_is_embedded

    def ends_embedded(self):
        return self._norm_minimizer_ends_embedded

    def is_admissible(self):
        return self._is_admissible
    
    def __str__(self):
        frac = '' if self._factor == 1 else self._factor
        vert_or_ray = 'Ray' if self._is_ray else 'Vertex' 
        if self.qtons_index() == None:
            mapped_from = 'not represented by any qtons'
        else:
            if self.is_embedded():
                mapped_from = 'mapped from embedded qtons with index {}'.format(self.qtons_index())
            else:
                mapped_from = 'mapped from immersed qtons with index {}'.format(self.qtons_index())

        if self._factor == 1:
            return '{} {}: represented by S_{},{} at {}, {}'.format(vert_or_ray, self.index(), self.genus(), self.num_boundary_comps(), self.coords(), mapped_from)
        else:
            return '{} {}: represented by ({})*S_{},{} at {}, {}'.format(vert_or_ray, self.index(), frac, self.genus(), self.num_boundary_comps(), self.coords(), mapped_from)

    def __repr__(self):
        return self.__str__()

    def detail(self):
        return self.__str__()

class Ray(NBVertex):
    def __init__(self, index, qtons_index, coords, TN_wrapper):
        NBVertex.__init__(self, index, qtons_index, coords, TN_wrapper)
        self._is_ray = True

class DualVertex(Vertex):
    def __init__(self, index, coords, dual_face):
        Vertex.__init__(self, index, coords)
        self._dual_face = dual_face

    def dual_face(self):
        return self._dual_face

class Facet(object):
    def __init__(self, index, dim, vertices, rays):
        self._index = index
        self._dim = dim
        self._vertices = vertices
        self._rays = rays

    def index(self):
        return self._index

    def dim(self):
        return self._dim

    def vertices(self):
        return self._vertices

    def rays(self):
        return self._rays

    
    def __str__(self):
        return 'facet of dimension {} with vertices:\n{}'.format(self.dim(),self.vertices())

    def __repr__(self):
        return self.__str__()










import tkinter as tk
from tkinter import ttk


class AutoScrollbar(ttk.Scrollbar):
    """ A scrollbar that hides itself if it's not needed. Works only for grid geometry manager """
    def set(self, lo, hi):
        if float(lo) <= 0.0 and float(hi) >= 1.0:
            self.grid_remove()
        else:
            self.grid()
            ttk.Scrollbar.set(self, lo, hi)

    def pack(self, **kw):
        raise tk.TclError('Cannot use pack with the widget ' + self.__class__.__name__)

    def place(self, **kw):
        raise tk.TclError('Cannot use place with the widget ' + self.__class__.__name__)



#-*-python-*-
from __future__ import print_function
from math import gcd as GCD

import sys
try:
    sys.path.remove('/Applications/Regina.app/Contents/MacOS/python')
except ValueError:
    pass
import regina

import snappy

from tnorm.kernel.simplicial import get_face_map_to_C2, get_quad_map_to_C2, H2_as_subspace_of_C2, qtons_image_in_C2
from tnorm.kernel.boundary import inward_oriented_bdy, outward_oriented_bdy
from tnorm.kernel.euler import solve_lin_gluing_eq, euler_char_
from tnorm.kernel.homology import _map_to_H2, _map_to_H1bdy
from tnorm.kernel.matrices import peripheral_curve_mats, oriented_quads_mat
from tnorm.utilities.regina_helpers import regina_to_sage_int
from tnorm.kernel.embedded import is_embedded, ends_embedded
from tnorm.norm_ball import TNormBall, DualNormBall, NBVertex, Ray, DualVertex
from tnorm.utilities.cached_prop import cached_property
from tnorm.kernel.peripheral import periph_basis_intersections, periph_basis_connected
from tnorm.utilities.sage_types import Matrix, vector, VectorSpace, Polyhedron, RR, QQ
#preparser(False)


class TN_wrapper(object):
	"""
	A class for working with Thurston Norm on H2(M,bdy(M)).

	sage: W = TN_wrapper(SnapPeaTri)

	SnapPeaTri can be any of the following:
	(1) A snappy Manifold or Triangulation
	(2) a regina.SnapPeaTriangulation triangulation
	(3) Any string argument that snappy.Manifold() can take, e.g.,

	sage: W = TN_wrapper('L6a5')
	sage: W = TN_wrapper('/path/to/snappy/file.tri')
	sage: W = TN_wrapper(string)       # where string is the triangulation encoded as a string (i.e., contents of a tri file)
	
	"""
	def __init__(self, manifold, qtons=None, quiet=False, tracker=False, allows_non_admissible=False, force_simplicial_homology=False):
		
		self._triangulation = None
		if isinstance(manifold,str):
			self._manifold = snappy.Manifold(manifold)
		elif isinstance(manifold, regina.engine.SnapPeaTriangulation):
			self._manifold = snappy.Manifold(manifold.snapPea())
			self._triangulation = manifold
		elif isinstance(manifold, regina.engine.Triangulation3):
			self._manifold = snappy.Manifold(manifold.snapPea())
			self._triangulation = manifold
		elif isinstance(manifold, snappy.Manifold):
			self._manifold = manifold
		elif isinstance(manifold, snappy.Triangulation):
			self._manifold = manifold

		for c in self._manifold.cusp_info():
			if c.is_complete == False:
				self._manifold = self._manifold.filled_triangulation()
				break
		
		self._QUIET = quiet
		self._force_simplicial_homology = force_simplicial_homology
		self._num_cusps = self._manifold.num_cusps()
		if self._num_cusps != 0:
			try: 
				L = self._manifold.link()
				self._knows_link_complement = True
			except ValueError:
				self._knows_link_complement = False
			#if self._knows_link_complement and bdy_H1_basis == 'natural':
			#	self._bdy_H1_basis = 'natural'
			#else:
			#	if self._manifold.verify_hyperbolicity()[0]:
			#		self._manifold.set_peripheral_curves('shortest')
			#		self._bdy_H1_basis = 'shortest'

			self._triangulation = regina.SnapPeaTriangulation(self._manifold._to_string())
			self._angle_structure = solve_lin_gluing_eq(self._triangulation)
			self._peripheral_curve_mats = peripheral_curve_mats(self._manifold, self._triangulation)
			self._manifold_is_closed = False


			# check to make sure peripheral basis curves actually give a basis
			check, message = periph_basis_intersections(self)
			assert check, message
			
			# check to make sure each peripheral basis curve is connected (extra trivial loops
			# will mess up Euler char calculations).
			check, message = periph_basis_connected(self)
			assert check, message

		else:
			if self._triangulation == None:
				self._triangulation = regina.Triangulation3(self._manifold._to_string())
				for _ in range(10):
					self._triangulation.intelligentSimplify()
			self._manifold_is_closed = True
			#self._bdy_H1_basis = None
		if not self._triangulation.isOriented():
			self._triangulation.orient()
		self._qtons = qtons
		self._tkr = False
		self._allows_non_admissible = allows_non_admissible
		self._is_fibered = 'unknown'
		self._betti_number = self._triangulation.homologyH1().rank()

		# qtons memoization caches-----
		self._euler_char = {}
		self._map_to_ball = {}
		self._map_to_H2 = {}
		self._num_boundary_comps = {}
		self._over_facet = {}
		self._is_norm_minimizing = {}
		self._is_admissible = {}
		self._qtons_image_in_C2 = {}
		self._num_H1bdy_comps = {}
		self._has_mixed_bdy = {}
		self._is_embedded = {}
		self._ends_embedded = {}
		self._oriented_quads_mat = {}

		if self._manifold.num_cusps() > 0:
			self._boundary_slopes = {}
			self._spinning_slopes = {}
			self._map_to_H1bdy = {}

		if not self._QUIET:		
			print('Enumerating quad transversely oriented normal surfaces (qtons)... ', end='')
			try:
				sys.stdout.flush()
			except AttributeError:
				pass

		if tracker:
			self._tkr = regina.ProgressTracker()

		# compute transversely oriented normal surfaces
		self._qtons = self.qtons()

		# name the surfaces by their index in the NormalSurfaces list
		for i in range(self._qtons.size()):
			self._qtons.surface(i).setName(str(i))

		if not self._QUIET:
			print('Done.')
			try:
				sys.stdout.flush()
			except AttributeError:
				pass

		if self._betti_number > self._num_cusps:
			self._has_internal_homology = True
		else:
			self._has_internal_homology = False

		# if the manifold has internal homology or force_simplicial_homology==True then we need to use simplicial homology.
		if self._has_internal_homology or self._force_simplicial_homology:
			
			self._uses_simplicial_homology = True

			if not self._QUIET:
				print('computing simplicial homology...',end='')
				try:
					sys.stdout.flush()
				except AttributeError:
					pass

			self._face_map_to_C2 = get_face_map_to_C2(self._triangulation)
			self._quad_map_to_C2 = get_quad_map_to_C2(self._triangulation, self._face_map_to_C2)
			H2_basis_in_C2, P, qtons_image = H2_as_subspace_of_C2(self, self._face_map_to_C2, self._quad_map_to_C2)
			self._project_to_im_del3 = P
			self._qtons_image_in_C2 = {i:qtons_image[i] for i in range(len(qtons_image))}
			assert len(H2_basis_in_C2) == self._betti_number, self._manifold.name()+', force_simplicial_homology={}'.format(self._force_simplicial_homology)
				# raise an error if the qtons do not generate H2. This should only happen for a knot in a 
				# rational homology sphere, with force_simplicial_homology=True.
			I = Matrix.identity(self._betti_number)
			B = Matrix(H2_basis_in_C2).transpose()
			A = B.solve_left(I)
			self._map_H2_to_standard_basis = A

			if not self._QUIET:
				print('Done.')
				try:
					sys.stdout.flush()
				except AttributeError:
					pass
		else:
			self._uses_simplicial_homology = False

	# need to make this work for qtons without index. I.e., a linear combo of vertex normal surfaces.
	
	def manifold(self):
		return self._manifold

	def triangulation(self):
		return self._triangulation

	def num_cusps(self):
		return self._num_cusps

	def knows_link_complement(self):
		return self._knows_link_complement

	#def bdy_H1_basis(self):
	#	return self._bdy_H1_basis

	def allows_non_admissible(self):
		return self._allows_non_admissible

	def betti_number(self):
		return self._betti_number

	def is_fibered(self):
		return self._is_fibered

	def has_internal_homology(self):
		return self._has_internal_homology

	def uses_simplicial_homology(self):
		return self._uses_simplicial_homology

	def manifold_is_closed(self):
		return self._manifold_is_closed


	def _simplicial_map_to_H2(self,qtons):
		if self.uses_simplicial_homology():
			try:
				ind = int(qtons)
			except TypeError:
				ind = int(qtons.name())
			if not ind in self._map_to_H2:
				s_H2 = self._map_H2_to_standard_basis*(self._qtons_image_in_C2[ind]-self._project_to_im_del3*self._qtons_image_in_C2[ind])
				self._map_to_H2[ind] = s_H2
			
			return self._map_to_H2[ind]
		else:
			return None

	def qtons(self):
		"""
		Enumerate vertex quad transversely oriented normal surfaces for the ideal triangulation self.triangulation()
		This is cached in self._qtons, so calls to this function after the initial call (when TN_wrapper() is 
		instantiated) will be fast. 

		Returns a regina.NormalSurfaces object.
		"""
		if self._qtons == None:
			if self._tkr:
				if self.allows_non_admissible():
					return regina.NormalSurfaces.enumerate(self.triangulation(),regina.NS_ORIENTED_QUAD,regina.NS_IMMERSED_SINGULAR,regina.NS_ALG_DEFAULT,self._tkr)
				else:
					return regina.NormalSurfaces.enumerate(self.triangulation(),regina.NS_ORIENTED_QUAD,regina.NS_VERTEX,regina.NS_ALG_DEFAULT,self._tkr)
			else:
				if self.allows_non_admissible():
					return regina.NormalSurfaces.enumerate(self.triangulation(),regina.NS_ORIENTED_QUAD,regina.NS_IMMERSED_SINGULAR,regina.NS_ALG_DEFAULT)
				else:
					return regina.NormalSurfaces.enumerate(self.triangulation(),regina.NS_ORIENTED_QUAD,regina.NS_VERTEX,regina.NS_ALG_DEFAULT)
		else:
			return self._qtons


	def show_progress(self):
		print(self._tkr.percent())
		try:
			sys.stdout.flush()
		except AttributeError:
			pass


	def euler_char(self,qtons):
		"""
		Return the Euler characteristic of the normal surface. 
		Argument can be a surface (Regina oriented spun normal surface) or the index of a surface.

		sage: W.euler_char(3)    # return Euler characteristic of the surface with index 3, i.e, W.OrientedNormalSurfaceList.surface(3).
		-17

		"""
		try:
			ind = int(qtons)
		except TypeError:
			ind = int(qtons.name())
			
		if not ind in self._euler_char:
			s = self.qtons().surface(ind)
			if not self.manifold_is_closed():
				ec = euler_char_(s,self._angle_structure)
			else:
				ec = regina_to_sage_int(s.eulerChar())
			self._euler_char[ind] = ec
		
		return self._euler_char[ind]

	def genus(self,qtons):
		"""
		Return the genus. Argument can be a surface (regina oriented
		spun normal surface) or the index of a surface.

		sage: W.genus(3)    # return genus of the surface with index 3, i.e, W.qtons().surface(3).
		1
		"""
		return (2-self.euler_char(qtons)-self.num_boundary_comps(qtons))/2

	def num_boundary_comps(self,qtons):
		"""
		Return the number of boundary components after mapping to H1(bdy m). This will be the number
		of boundary components after cancelling oppositely oriented pairs (i.e., a 3-puntured sphere
		will become a once-punctured torus if two of the punctures are on the cusps and boundary curves
		with opposite orientations). Argument can be a surface (regina oriented spun normal surface) or 
		the index of a surface.

		sage: W.numBoundaryComps(3)    # return num of boundary components of surface with index 3, i.e, W.OrientedNormalSurfaceList.surface(3).
		2

		"""	
			
		try:
			ind = int(qtons)
			s = self.qtons().surface(ind)
		except TypeError:
			ind = int(qtons.name())
			s = qtons

		if not ind in self._num_boundary_comps:
			if self.manifold_is_closed():
				self._num_boundary_comps[ind] = 0
			else:
				nbc = 0
				pos, neg = self.boundary_slopes(ind).values()
				for slope in pos:
					nbc += gcd(slope[0],slope[1])
				for slope in neg:
					nbc += gcd(slope[0],slope[1])
				self._num_boundary_comps[ind] = nbc

		return self._num_boundary_comps[ind]	

	def boundary_slopes(self,qtons):
		"""
		Return the positive and negative boundary slopes of the given surface in H1(bdy(M)), with respect to the basis of peripheral curves given by SnapPy. 
		Argument can be a surface (regina quad transverely oriented normal surface) or the index of a surface. Returns as a tuple of
		the form (pos_slopes, neg_slopes), where pos_slopes is a list such that pos_slopes[i] is positive boundary slope on cusp i,
		and similarly for neg_slopes.

		sage: W.boundary_slopes(3)    # return positive and negative boudnary of the surface with index 3, i.e, W.OrientedNormalSurfaceList.surface(3).
		([(0, 0), (-3, 1)], [(3, -1), (0, 0)])
		"""
		if self.manifold_is_closed():
			return []
		else:
			try:
				ind = int(qtons)
			except TypeError:
				ind = int(qtons.name())
	
			if not ind in self._boundary_slopes:
				s = self.qtons().surface(ind)
				out_bdy = outward_oriented_bdy(s, self._peripheral_curve_mats, self.oriented_quads_mat(s), False)
				in_bdy = inward_oriented_bdy(s, self._peripheral_curve_mats, self.oriented_quads_mat(s), False)
				self._boundary_slopes[ind] = {'outward':out_bdy, 'inward':in_bdy}
			
			return self._boundary_slopes[ind]

	def oriented_quads_mat(self, qtons):
		try:
			ind = int(qtons)
		except TypeError:
			ind = int(qtons.name())

		if not ind in self._oriented_quads_mat:
			s = self.qtons().surface(ind)
			self._oriented_quads_mat[ind] = oriented_quads_mat(s)
		
		return self._oriented_quads_mat[ind]

	def is_embedded(self,qtons):
		"""Return True if the given qtons can be normally isotoped to be embedded.
		"""
		try:
			ind = int(qtons)
		except TypeError:
			ind = int(qtons.name())

		if not ind in self._is_embedded:
			s = self.qtons().surface(ind)
			self._is_embedded[ind] = is_embedded(s,self)
		
		return self._is_embedded[ind]

	def ends_embedded(self,qtons):
		"""Return True if the ends of the given qtons can be normally isotoped to be embedded.
		"""
		if self.manifold_is_closed():
			return True

		try:
			ind = int(qtons)
		except TypeError:
			ind = int(qtons.name())

		if not ind in self._ends_embedded:
			s = self.qtons().surface(ind)
			self._ends_embedded[ind] = ends_embedded(s,self)
		
		return self._ends_embedded[ind]		

	def has_mixed_bdy(self, qtons):
		if self.manifold_is_closed():
			return False
		else:
			try:
				ind = int(qtons)
			except TypeError:
				ind = int(qtons.name())
	
			if not ind in self._has_mixed_bdy:
				pos_bdy, neg_bdy = self.boundary_slopes(ind).values()
				if Matrix(pos_bdy).norm() == 0 or Matrix(neg_bdy).norm() == 0:
					self._has_mixed_bdy[ind] = False
				else:
					self._has_mixed_bdy[ind] = True

			return self._has_mixed_bdy[ind]


	def spinning_slopes(self,qtons):
		"""
		Return the positive and negative boundary slopes of the given surface in H1(bdy(M)), with respect to the basis of peripheral curves given by SnapPy.
		Argument can be a surface (regina quad transverely oriented normal surface) or the index of a surface.

		sage: W._map_to_H1bdy(3)    # return image in H1(bdy(M)) of the surface with index 3, i.e, W.OrientedNormalSurfaceList.surface(3).
		[(-1, 0), (0, 1), (-1, 0)]
		"""
		if self.manifold_is_closed():
			return []
		else:
			try:
				ind = int(qtons)
			except TypeError:
				ind = int(qtons.name())
	
			if not ind in self._spinning_slopes:
				s = self.qtons().surface(ind)
				out_bdy = outward_oriented_bdy(s, self._peripheral_curve_mats, self.oriented_quads_mat(s), True)
				in_bdy = inward_oriented_bdy(s, self._peripheral_curve_mats, self.oriented_quads_mat(s), True)
				self._spinning_slopes[ind] = out_bdy, in_bdy
			
			return self._spinning_slopes[ind]



	def H1bdy_slopes(self, qtons):
		
		if self.manifold_is_closed():
			return []
		else:
			try:
				ind = int(qtons)
			except TypeError:
				ind = int(qtons.name())
	
			if not ind in self._map_to_H1bdy:
				s = self.qtons().surface(ind)
				slopes = _map_to_H1bdy(s, self)
				self._map_to_H1bdy[ind] = slopes
			
			return self._map_to_H1bdy[ind]	

	def num_H1bdy_comps(self, qtons):
		try:
			ind = int(qtons)
			s = self.qtons().surface(ind)
		except TypeError:
			ind = int(qtons.name())
			s = qtons

		if not ind in self._num_H1bdy_comps:
			nbc = 0
			slopes = self.H1bdy_slopes(ind)
			for slope in slopes:
				nbc += gcd(slope[0],slope[1])
			self._num_H1bdy_comps[ind] = nbc

		return self._num_H1bdy_comps[ind]	


	def simplicial_class(self,qtons):
		""" Return the vector in the dimension 2 chain group C2 corresponding to the qtons surface. The i^th coordinate
			of this vector corresponds to the number of triangles of index i, as indexed by Regina, with orientation positive
			if the transverse orientation of the surface points toward the tetrahedron triangle(i).front(). 
		"""
		if not self.uses_simplicial_homology():
			return None
		else:
			try:
				ind = int(qtons)
			except TypeError:
				ind = int(qtons.name())
	
			if not ind in self._qtons_image_in_C2:
				c = qtons_image_in_C2(self,self._quad_map_to_C2)
				self._qtons_image_in_C2[ind] = c

			return self._qtons_image_in_C2[ind]

	def over_facet(self, qtons, as_string=False):
		"""Returns the facet of the norm ball that the qtons surface is above. E.g., <0 1 5> corresponds to the facet 
		with vertices 0, 1, and 5. 
		"""
		try:
			ind = int(qtons)
		except TypeError:
			ind = int(qtons.name())

		if ind in self._over_facet:
			f = self._over_facet[ind]
		else:
			v = self.map_to_H2(ind)
			f = over_facet_(v,self.norm_ball.polyhedron())
			self._over_facet[ind] = f
		if as_string:
			return None if f==None else '<{}>'.format(' '.join([str(v.index()) for v in f.vertices()]))
		else:
			return f

	def is_norm_minimizing(self, qtons):
		"""Return True if the given qtons surface is Thurston norm minimizing.
		"""
		try:
			ind = int(qtons)
		except TypeError:
			ind = int(qtons.name())

		if not ind in self._is_norm_minimizing:
			v = self.map_to_H2(ind)
			if not self.norm_ball.polyhedron().interior_contains(v):
				self._is_norm_minimizing[ind] = True
				return True
			else:
				self._is_norm_minimizing[ind] = False

		return self._is_norm_minimizing[ind]

	def is_admissible(self, qtons):
		try:
			ind = int(qtons)
		except TypeError:
			ind = int(qtons.name())

		if not ind in self._is_admissible:
			s = self.qtons().surface(ind)
			mat = oriented_quads_mat(s)
			for row in mat:
				nonzero = [i for i in range(0,6,2) if ( row[i]!=0 or row[i+1]!=0 )]
				if len(nonzero) > 1:
					self._is_admissible[ind] = False
					return False
			self._is_admissible[ind] = True

		return self._is_admissible[ind]

	def map_to_ball(self, qtons):
		"""
		Return the image of the given surface in H2(M, bdy(M)), then divide by Euler characteristic. Argument can be a surface (regina oriented
		spun normal surface) or the index of a surface.

		"""
		try:
			ind = int(qtons)
		except TypeError:
			ind = int(qtons.name())

		if not ind in self._map_to_ball:
			image_in_H2 = self.map_to_H2(ind)
			if self.euler_char(ind) < 0:
				mtb = tuple([QQ(i)/(-self.euler_char(ind)) for i in image_in_H2])
				self._map_to_ball[ind] = mtb
				return mtb
			else:
				self._map_to_ball[ind] = float('inf')

		return self._map_to_ball[ind]


	def map_to_H2(self, qtons):
		"""
		Return the image of the given surface in H2(M, bdy(M)). Argument can be a surface (regina quad transversely oriented
		normal surface) or the index of a surface.

		sage: W.mapToH2(3)    # return the image in H2(M,bdy(M)) of the surface with index 3, i.e, W.OrientedNormalSurfaceList.surface(3).
		(0,1,0)

		"""
		if not self.uses_simplicial_homology():
			try:
				ind = int(qtons)
			except TypeError:
				ind = int(qtons.name())
	
			if not ind in self._map_to_H2:
				s = self.qtons().surface(ind)
				self._map_to_H2[ind] = _map_to_H2(s, self)
			
			return self._map_to_H2[ind]
		else:
			return self._simplicial_map_to_H2(qtons)
		
	def _image_in_H2M(self):
		image = []
		for i in range(self.qtons().size()):
			pt = self.map_to_H2(i)
			if not pt.is_zero():
				ec = self.euler_char(i)
				#if (pt,ec) not in [tup[1:] for tup in image]:
				image.append((i,pt,ec))
		return image

	@cached_property
	def _norm_ball_points(self):
		"""
		Return the points in H2(M,bdy M) which are images of (normalized) vertices of the oriented spun
		normal surface projective solution space with negative Euler characteristic. The convex hull of 
		these points is the norm ball.
		"""
		points0 = dict()
		points = dict()
		rays = dict()  # if the Euler char of s is 0 we can't normalize. This means the norm ball is
		               # infinite in the direction of map_to_H2(s), so it contains the ray (0,map_to_H2(s)).
		image_w_euler = self._image_in_H2M()


		for (i,pt,ec) in image_w_euler:
			if ec < 0:
				normalized = tuple([coord/(-ec) for coord in pt]) # divide by euler char to normalize

#				if not self.uses_simplicial_homology:
				if normalized not in points0:
					points0[normalized] = []
					points0[normalized].append((i,pt,ec))
				else:
					points0[normalized].append((i,pt,ec))

			elif ec == 0:
				rays[tuple(pt)] = i

		for nzd in points0:

			min_ec = min(-rep[2] for rep in points0[nzd])
			reps = [rep[0] for rep in points0[nzd] if -rep[2]==min_ec] # restrict to norm minimizing reps

			if self._force_simplicial_homology:
				# order reps by L^1 norm
				reps = sorted(reps,key=lambda x: sum(self.simplicial_class(x).apply_map(abs)))
				points[nzd] = reps[0] # get the index of the rep with minimal L^1 norm
			else:
				good_ends = [rep for rep in reps if self.ends_embedded(rep)]
				if len(good_ends) != 0:
					reps = good_ends # if possible, choose from reps with embedded ends
				embedded = [rep for rep in reps if self.is_embedded(rep)]
				if len(embedded) != 0:
					reps = embedded # if possible, choose from reps that are embedded
				reps = sorted(reps,key=lambda x: self.num_boundary_comps(x))
				points[nzd] = reps[0] # get the index of the rep with the minimal number of bdy components

		return points, rays

	@cached_property
	def norm_ball(self):
		"""
		Return the Thurston norm ball.
		"""
		if not self._QUIET:
			print('Computing Thurston norm unit ball... ', end='')
			try:
				sys.stdout.flush()
			except AttributeError:
				pass
		pts_dict, rays_dict = self._norm_ball_points

		### If M is not hyperbolic then rays_dict should be non-empty, and the norm ball should be non-compact. 
		### The below attempts to compute the non-compact norm ball, but it may be wrong. This is because 
		### (it seems) some surfaces may not be realized as spun normal surfaces.

		if len(rays_dict) != 0:
			V=VectorSpace(RR,self.betti_number())
			ray_span=V.span([V(r) for r in rays_dict])
			keys = [p for p in pts_dict]
			for p in keys:
				if V(p) in ray_span:
					_ = pts_dict.pop(p)


			polyhedron = Polyhedron(vertices=Matrix(pts_dict.keys()), rays=Matrix(rays_dict.keys()), base_ring=QQ, backend='cdd')
			rays = tuple([(rays_dict[tuple(i*vector(v))],i*vector(v)) for v in polyhedron.lines_list() for i in [1,-1]])
			Rays = [Ray(i, rays[i][0], rays[i][1], self) for i in range(len(rays))]
			vertices = tuple([(pts_dict[tuple(v)],vector(v)) for v in polyhedron.vertices_list() if not vector(v).is_zero()])
			Vertices = [NBVertex(i, vertices[i][0], vertices[i][1], self) for i in range(len(vertices))]

		else:
			polyhedron = Polyhedron(vertices=Matrix(pts_dict.keys()), base_ring=QQ)
			vertices = tuple([(pts_dict[tuple(v)],vector(v)) for v in polyhedron.vertices_list()])
			
			Vertices = [NBVertex(i,vertices[i][0],vertices[i][1], self) for i in range(len(vertices))]
			Rays = []

		ball = TNormBall(Vertices, Rays, polyhedron)
		if self.num_cusps() == 1 and self.betti_number() == 1:
			M = self.manifold().copy()
			(p,q) = self.manifold().homological_longitude()
			elem_divs = [div for div in M.homology().elementary_divisors() if div != 0]
			M.dehn_fill((p,q))
			filled_elem_divs = [div for div in M.homology().elementary_divisors() if div != 0]
			b = 1
			for div in elem_divs:
				b *= div
			for div in filled_elem_divs:
				b /= div
			A = self.manifold().alexander_polynomial()
			A_norm = QQ(A.degree() - 1)
			try:
				v=ball.vertices()[0]

			except IndexError:
				pass

			### below needs to be fixed. Currently does not account for virtual fibers.
			if A_norm <= 0 or not A.is_monic():
				self._is_fibered = False
				ball._confirmed = True

			elif len(ball.vertices())==0:
				self._is_fibered = True
				polyhedron = Polyhedron(vertices=[[A_norm],[-A_norm]], base_ring=QQ)
				Vertices = [NBVertex(0, None, (-1/A_norm,), self, (b, -A_norm, {'outward':[(0,1)],'inward':[(0,0)]})), NBVertex(1, None, (1/A_norm,), self, (b, -A_norm, {'outward':[(0,-1)],'inward':[(0,0)]}))]
				ball = TNormBall(Vertices, Rays, polyhedron)
				ball._confirmed = True
			elif A_norm == abs(v.euler_char()):
				assert self.num_H1bdy_comps(v.qtons_index()) == b
				self._is_fibered = 'unknown'
				ball._confirmed = True
			elif A_norm < abs(v.euler_char()):
				#M = self.manifold().copy()
				#M.dehn_fill(M.homological_longitude())
				Mf = M.filled_triangulation()
				T = regina.Triangulation3(Mf._to_string())
				boo = T.intelligentSimplify()
				ns = regina.NormalSurfaces.enumerate(T,regina.NS_QUAD,regina.NS_VERTEX,regina.NS_ALG_DEFAULT)
				non_trivial = [i for i in range(ns.size()) if ns.surface(i).isOrientable()]
				if len(non_trivial)>1:
					non_trivial = [i for i in non_trivial if ns.surface(i).cutAlong().isConnected()]
				if len(non_trivial) == 0: ## something is wrong!
					print('Warning: failed to confirm that norm ball is correct (error: len(non_trivial)==0, culprit:M={}.'.format(self.manifold().name()))
					ball._confirmed = False
				elif len(non_trivial) >= 1:
					genus = min([(2-regina_to_sage_int(ns.surface(i).eulerChar()))/2 for i in non_trivial])
					if 2*genus - 2 + b == abs(v.euler_char()):
						self._is_fibered = False
						ball._confirmed = True
					elif 2*genus - 2 + b == A_norm:
						multiplier = 1/QQ(abs(v.euler_char())/A_norm)
						if self.uses_simplicial_homology() == True:
							simplicial_class = multiplier*self.simplicial_class(v.qtons_index())
						else:
							simplicial_class = None
						self._is_fibered == True
						polyhedron = Polyhedron(vertices=[[A_norm],[-A_norm]], base_ring=QQ)
						Vertices = [NBVertex(0, None, (-1/A_norm,), self, (b, -A_norm, {'outward':[(0,1)],'inward':[(0,0)]})),NBVertex(1, None, (1/A_norm,), self, (b, -A_norm, {'outward':[(0,-1)],'inward':[(0,0)]}))]
						ball = TNormBall(Vertices, Rays, polyhedron)
						ball._confirmed = True
					else:
						print('Warning: failed to confirm that norm ball is correct (error: 2g-1!=A_norm or abs(X(S)), culprit:M={}.'.format(self.manifold().name()))
						ball._confirmed = False
			else:
				# if we are here, then something is wrong, and the following will cause an error to be thrown.
				assert A_norm <= abs(v.euler_char()) # (this should never happen)
		if not self._QUIET:
			print('Done.')
			try:
				sys.stdout.flush()
			except AttributeError:
				pass
		return ball

	@cached_property
	def dual_norm_ball(self):
		B = self.norm_ball
		P = B.polyhedron()
		poly_vertices = []
		vertices = []
		for i in range(len(P.faces(P.dim()-1))):
			face = P.faces(P.dim()-1)[i]
			e = face.as_polyhedron().equations_list()[0]
			v = -vector(e[1:])/(e[0])
			poly_vertices.append(v)
			vertices.append(DualVertex(i, v, B.facets(P.dim()-1)[i]))
		P_dual = Polyhedron(vertices=poly_vertices, base_ring=QQ)
		dual_ball = DualNormBall(vertices, P_dual)
		return dual_ball

	@cached_property
	def qtons_info(self):
		if not self._QUIET:
			print('Analyzing quad transversely oriented normal surfaces... ', end='')
			try:
				sys.stdout.flush()
			except AttributeError:
				pass
		qtons_info_dict = dict([(i,{}) for i in range(self.qtons().size())])
		for i in range(self.qtons().size()):
			qtons_info_dict[i]['image_in_H2'] = self.map_to_H2(i)
			qtons_info_dict[i]['euler_char'] = self.euler_char(i)
			qtons_info_dict[i]['boundary_slopes'] = self.boundary_slopes(i)
			qtons_info_dict[i]['is_embedded'] = self.is_embedded(i)
			qtons_info_dict[i]['num_boundary_comps'] = self.num_boundary_comps(i)
			qtons_info_dict[i]['genus'] = self.genus(i)
			qtons_info_dict[i]['is_norm_minimizing'] = self.is_norm_minimizing(i)
			qtons_info_dict[i]['over_facet'] = self.over_facet(i,as_string=True)
		if not self._QUIET:
			print('Done.')
			try:
				sys.stdout.flush()
			except AttributeError:
				pass
		return qtons_info_dict

def check_subfaces(ray,f_poly):
	if f_poly.dim() == 0:
		return f_poly
	for f in f_poly.faces(f_poly.dim()-1):
		if not f.as_polyhedron().intersection(ray).is_empty():
			return check_subfaces(ray,f.as_polyhedron())
		else:
			continue
	return f_poly
			

def over_facet_(v, P):
	if v.is_zero():
		return None
	else:
		ray = Polyhedron(rays=[v]) # ray generated by v
		f_poly = check_subfaces(ray, P) # find the facet that v lies over, as a polyhedron

		# f_poly is the facet we want, but it is in the form of a polyhedron, not a facet of P. So we find the facet of P with the same vertices, and return that.
		for i in range(P.dim()):
			for f in P.faces(i):
				if sorted([vert.vector() for vert in f.vertices()]) == sorted([vert.vector() for vert in f_poly.vertices()]):
					return f

def lies_over_facet(v, P, facet):
	# find the lowest dimensional facet that v lies over
	v_facet = over_facet_(v, P)
	# check if v_facet found above is a subfacet of facet
	if set(v_facet.vertices()).issubset(set(facet.vertices())) and set(v_facet.lines()).issubset(set(facet.lines())):
		return True
	return False

def basis_over_facet(basis, P, facet):  # check if all vectors in basis lie over facet.
	for vec in basis:
		if not lies_over_facet(vec, P, facet):
			return False
	return True

def orthogonal_proj(v, rays):
	coeffs = [QQ(1)]
	for w in rays:
		w = vector(w)
		coeff = QQ(-(v.dot_product(w)/w.dot_product(w)))
		v = v + coeff*w
		coeffs.append(coeff)
	l = lcm([coeff.denominator() for coeff in coeffs])
	if l != 1:
		coeffs = [coeff*l for coeff in coeffs]
		v = v*l
	return v, coeffs

def abs_slopes(slopes):
	return [tuple([abs(int(c)) for c in s]) for s in slopes]

def int_slopes(slopes):
	return [tuple([int(c) for c in s]) for s in slopes]


def lcm(list_of_ints):
    L = list_of_ints
    LCM = 1
    for k in L:
        LCM = LCM*k//gcd(LCM,k)
    return LCM

def gcd(a,b):
    return abs(GCD(a,b))












from tnorm.utilities.sage_types import vector, Matrix
from tnorm.utilities.regina_helpers import regina_to_sage_int

"""
Let t be the number of tetrahedra for a triangulation T of M, and let v be the number of vertices (if M is cusped this is the
number of cusps). Then T has 4t/2=2t faces and which generate the dim-2 chain group C_2. A (possibly closed, not nec. embedded)
spun normal surface S is identified with an element of C_2, via an identification of quads with pairs of tetrahdera faces: an 
oriented quad q in a tetahedron t is isotopic to the pair of triangles of t (with outward tranverse orientation) that meet
at the edge that q misses, and which the transverse orientation of q points away from (see Segerman's "On spun normal and twisted
square surfaces"). By mapping all qtons to C2 via this scheme, we get a subspace K, which is the kernel of the boundary map 
from C2-->C1. We also get the image L of the boundary map C3-->C2 by taking the span of the images del3(t) for tetrahedra t in T.

Let dim(C2)=n, let dim(K)=k, and let dim(L)=l. let b_1,...,b_k be vectors in C2 forming a basis for K, and let c_1,...,c_l be 
vectors is C2 forming a basis for L. Let P be projection onto L. Then Q*b_i := b_i-P*b_i is in the orthogonal complement of L in K, as 
a subspace of C2. Let O=span{(Q*b_i)}_i. dim(O)=d should be the betti number of T, i.e., the dimension of H2 rel bdy.
Let {v_1,...,v_d} be a spanning set of vectors for O. Note that the v_i are not necesarily orthogonal, and there is no reason
to make them orthogonal, as we will map them to the stardand basis below.

Next find a map A:O-->QQ^d by solving A*B=Id_k where B=[v_1,...,v_d] is a matrix whose columns are the spanning vectors 
{v_1,...,v_d}. Now just map all normal surfaces down to H2 via A*Q. Note that the v_i come from cycles, so the basis elements 
for H2 are associated to normal surfaces.
"""


def standard_basis_vec(k, dim):
    return vector([0 if i != k else 1 for i in range(dim)])

def standard_basis_tup(k, dim):
    return tuple([0 if i != k else 1 for i in range(dim)])

def get_face_map_to_C2(T):
    tris = T.triangles()
    num_faces = len(tris)
    C2_face_map = {}
    for k in range(num_faces):
        t = tris[k]
        v = standard_basis_vec(k,num_faces)
        C2_face_map[(t.front().tetrahedron().index(),t.front().triangle())] = -v*(2*(t.front().triangle()%2)-1)
        C2_face_map[(t.back().tetrahedron().index(),t.back().triangle())] = v*(2*(t.front().triangle()%2)-1)

    return C2_face_map

def del3_matrix(T):
    rows = []
    C2_face_map = get_face_map_to_C2(T)
    for i in range(T.size()):
        row = vector([0 for _ in range(2*T.size())])
        for j in range(4):
            row += vector(C2_face_map[(i,j)])
        rows.append(row)
    mat = Matrix(rows).transpose()
    return mat


def get_quad_map_to_C2(T,C2_face_map):
    C2_quad_map = {}
    for i in range(T.size()):
        C2_quad_map[(i,0,False)] = C2_face_map[(i,0)]+C2_face_map[(i,1)]
        C2_quad_map[(i,0,True)] = C2_face_map[(i,2)]+C2_face_map[(i,3)]
        C2_quad_map[(i,1,False)] = C2_face_map[(i,0)]+C2_face_map[(i,2)]
        C2_quad_map[(i,1,True)] = C2_face_map[(i,1)]+C2_face_map[(i,3)]
        C2_quad_map[(i,2,False)] = C2_face_map[(i,0)]+C2_face_map[(i,3)]
        C2_quad_map[(i,2,True)] = C2_face_map[(i,1)]+C2_face_map[(i,2)]
    return C2_quad_map

#this maps a normal surface to C2
def map_qtons_to_C2(surface, C2_quad_map):
    T = surface.triangulation()
    image = vector([0 for i in range(2*T.size())])
    for i in range(T.size()):
        for j in range(3):
            for boo in [True,False]:
                image += regina_to_sage_int(surface.orientedQuads(i,j,boo))*C2_quad_map[(i,j,boo)]
    return image

# this returns vectors whose span is the subspace K of C2 in the notation above
def qtons_image_in_C2(TN_Wrapper, C2_quad_map): 
    W = TN_Wrapper
    image = []
    for i in range(W.qtons().size()):
        image.append(map_qtons_to_C2(W.qtons().surface(i),C2_quad_map))
    return image

def kernel_of_del2(TN_Wrapper, C2_quad_map):
    W = TN_Wrapper
    qtons_image = qtons_image_in_C2(W, C2_quad_map)
    return gram_schmidt(qtons_image)


# this returns vectors whose span is the subspace L of C2 in the notation above
def image_of_del3(TN_Wrapper, C2_face_map):
    W = TN_Wrapper
    T = W.triangulation()
    image = []
    for i in range(T.size()):
        tet = T.tetrahedron(i)
        bdy = 0
        for j in range(4):
            bdy += C2_face_map[(i,j)]
        image.append(bdy)
    return gram_schmidt(image)

# find qtons reps that realize longitudes.
def get_periph_basis_surfaces(TN_Wrapper):
    surfaces = []
    periph_basis_found = []
    W = TN_Wrapper
    if W.manifold().num_cusps() == 0:
        return []
    else:
        T = W.triangulation()
        num_cusps = W.manifold().num_cusps()
        betti = W.betti_number()
        periph_H2_std_vecs = [standard_basis_tup(k,num_cusps) for k in range(num_cusps)]
        for i in range(W.qtons().size()):
            H1bdy_slopes = W.H1bdy_slopes(i)
            periph_H2 = tuple([H1bdy_slopes[j][1] for j in range(len(H1bdy_slopes))])
            if periph_H2 in periph_H2_std_vecs:
                if periph_H2 not in periph_basis_found:
                    periph_basis_found.append(periph_H2)
                    surfaces.append(i)
        return surfaces

def get_nontrivial_closed(TN_Wrapper):
    W = TN_Wrapper
    return [i for i in range(W.qtons().size()) if W.qtons().surface(i).isCompact() and W.qtons().surface(i).cutAlong().isConnected()]


# returns vectors in C2 that form a basis for H2 (as a subspace of C2). This is O in the notation at the top.
def H2_as_subspace_of_C2(TN_Wrapper, C2_face_map, C2_quad_map):
    W = TN_Wrapper
    periph = get_periph_basis_surfaces(W)
    closed = get_nontrivial_closed(W)
    others = [i for i in range(W.qtons().size()) if ( not W.qtons().surface(i).isCompact() ) and ( i not in periph )]
    all_qtons = periph + closed + others
    qtons_image = qtons_image_in_C2(W,C2_quad_map)
    cycles = [qtons_image[i] for i in all_qtons]
    #K = gram_schmidt(cycles)
    L = image_of_del3(W, C2_face_map)

    proj = projection(L)
    non_orthog_H2_basis_in_C2 = find_spanning_subset([b-proj*b for b in cycles])

    return non_orthog_H2_basis_in_C2, proj, qtons_image

# same as gram_schmidt below, except we keep the original vectors instead of the orthogonalized vectors.
# So this just extracts the first d vectors in the list that span, where d=dim(span{vectors})
def find_spanning_subset(vectors): 
    dim = len(vectors[0])
    orthobasis = []
    spanning_subset = []
    for i in range(len(vectors)):
        orthobasis.append(vectors[i])
        spanning_subset.append(vectors[i])
        for j in range(0,len(orthobasis)-1):
            orthobasis[-1] -= ((vectors[i].dot_product(orthobasis[j]))/(orthobasis[j].dot_product(orthobasis[j])))*orthobasis[j]
        if orthobasis[-1].is_zero():
            orthobasis = orthobasis[:-1]
            spanning_subset = spanning_subset[:-1]
            
    return spanning_subset

# below implementation of gram-schmidt does not require vectors to be lin. ind. returns an orthogonal basis for 
# subspace generated by vectors. I.e., this returns span{vectors}.
def gram_schmidt(vectors): 
    dim = len(vectors[0])
    orthobasis = []
    for i in range(len(vectors)):
        orthobasis.append(vectors[i])
        for j in range(0,len(orthobasis)-1):
            orthobasis[-1] -= ((vectors[i].dot_product(orthobasis[j]))/(orthobasis[j].dot_product(orthobasis[j])))*orthobasis[j]
        if orthobasis[-1].is_zero():
            orthobasis = orthobasis[:-1]
            
    return orthobasis



def projection(vectors):
    A = Matrix(vectors).transpose()
    P = A*((A.transpose()*A).inverse())*A.transpose()
    return P


def del2_matrix(triangulation):
    bdy = []
    for tri in triangulation.triangles():
        row = [0] * triangulation.countEdges() # == countFaces(1)

        for i in range(3):
            edge_index = tri.edge(i).index()
            perm = tri.edgeMapping(i) # == tri.faceMapping(1,i)
            row[edge_index] += perm.sign()
        bdy.append(row)
    bdy = Matrix(bdy).transpose()
    return bdy








from tnorm.utilities.sage_types import Matrix
from tnorm.utilities.regina_helpers import get_oriented_quads, get_quads, in_cusp


### for given cusp, return the matrix that gives intersection data for longitude and meridian.
### The Matrix mat_l has rows corresponding to tetrahedra, each row has length 16: the jth entry is 
### (j (mod 4) - 1) face of the the boundary triangle at the floor(j/4) vertex of the tet. A positive
### entry indicates that the longitude is going into the triangle, and a negative entry indicates
### it is exiting. Similarly for mat_m (meridian).
###                          v0    |     v1    |    v2     |    v3
###                    _                                               _
###               t0  | f0 f1 f2 f3 f0 f1 f2 f3 f0 f1 f2 f3 f0 f1 f2 f3 |
###               t1  |                       .                         |
###               .   |                       .                         |
###               .   |                       .                         |
###               .   |                       .                         |
###               .   |                       .                         |
###               tk  |_                                               _|

def peripheral_curve_mats(M,T):
    mats = []
    for cusp in range(T.countCusps()):
        tets = T.size()
        try:
            # for newer versions of SnapPy
            mat = M._get_cusp_indices_and_peripheral_curve_data()[1]
        except AttributeError:
            # for older versions of SnapPy
            mat = M._get_peripheral_curve_data()
        mat_l=Matrix([[mat[i][j]*int(in_cusp(T,i//4,j//4,cusp)) for j in range(16)] for i in range(2,4*tets,4)])
        mat_m=Matrix([[mat[i][j]*int(in_cusp(T,i//4,j//4,cusp)) for j in range(16)] for i in range(0,4*tets,4)])
        mats.append((mat_l,mat_m))
    return mats



### Quad coordinate of the surface as a Matrix. Row i corresponds to tetrahedron i, and each row has
### six entries corresponding to the six quad types: (0,1), (2,3), (0,2), (1,3), (0,3), (1,2) in that order.

def oriented_quads_mat(qtons):
    s = qtons
    T = s.triangulation()
    return Matrix([[get_oriented_quads(s,i,j,k) for (j,k) in [(0,1),(2,3),(0,2),(1,3),(0,3),(1,2)]] for i in range(T.size())])


### Quad coordinate of the surface as a Matrix. Row i corresponds to tetrahedron i, and each row has
### three entries corresponding to the three unoriented quad types: (0,1), (0,2), (0,3), in that order.
def quads_mat(spun_surface):
    s = spun_surface
    T = s.triangulation()
    return Matrix([[get_quads(s,i,j,k) for (j,k) in [(0,1),(0,2),(0,3),]] for i in range(T.size())])



### Same form as above matrix, but only picks up quads that have a boundary arc isotopic into the given
### cusp
def quad_bdy_mat(qtons, cusp):
    s = qtons
    T = s.triangulation()
    quads_mat = []
    for i in range(T.size()):
        row = []
        for (j,k) in [(0,1),(2,3),(0,2),(1,3),(0,3),(1,2)]:
            if ( in_cusp(T,i,j,cusp) or in_cusp(T,i,k,cusp) ):
                row.append(get_oriented_quads(s,i,j,k))
            else:
                row.append(0)
        quads_mat.append(row)
    return Matrix(quads_mat)


### Matrix that encodes how curves intersect with quad types, restricted to intersections
### that contribute to the positive boundary (see Cooper--Tillmann--Worden Fig 5).def pos_intx_matrix():

POS_INTERSECTION_MAT = Matrix([[0, 0, 0, 0, 0, 0],
                               [1,0, 0, 0, 0, 0],
                               [0, 0, 1,0, 0, 0],
                               [0, 0, 0, 0, 1,0],
                               [1,0, 0, 0, 0, 0],
                               [0, 0, 0, 0, 0, 0],
                               [0, 0, 0, 0,0, 1],
                               [0, 0,0, 1, 0, 0],
                               [0, 0, 1,0, 0, 0],
                               [0, 0, 0, 0,0, 1],
                               [0, 0, 0, 0, 0, 0],
                               [0, 1, 0, 0, 0, 0],
                               [0, 0, 0, 0, 1,0],
                               [0, 0,0, 1, 0, 0],
                               [0, 1, 0, 0, 0, 0],
                               [0, 0, 0, 0, 0, 0]])

### Matrix that encodes how curves intersect with quad types, restricted to intersections
### that contribute to the negative boundary (see Cooper--Tillmann--Worden Fig 5).
NEG_INTERSECTION_MAT = Matrix([[0, 0, 0, 0, 0, 0],
                               [0,-1, 0, 0, 0, 0],
                               [0, 0, 0,-1, 0, 0],
                               [0, 0, 0, 0, 0,-1],
                               [0,-1, 0, 0, 0, 0],
                               [0, 0, 0, 0, 0, 0],
                               [0, 0, 0, 0,-1, 0],
                               [0, 0,-1, 0, 0, 0],
                               [0, 0, 0,-1, 0, 0],
                               [0, 0, 0, 0,-1, 0],
                               [0, 0, 0, 0, 0, 0],
                               [-1, 0, 0, 0, 0, 0],
                               [0, 0, 0, 0, 0,-1],
                               [0, 0,-1, 0, 0, 0],
                               [-1, 0, 0, 0, 0, 0],
                               [0, 0, 0, 0, 0, 0]])
### Matrix that encodes how curves intersect with quad types (see Cooper--Tillmann--Worden Fig 5).
INTERSECTION_MAT = Matrix([[0, 0, 0, 0, 0, 0],
                           [1,-1, 0, 0, 0, 0],
                           [0, 0, 1,-1, 0, 0],
                           [0, 0, 0, 0, 1,-1],
                           [1,-1, 0, 0, 0, 0],
                           [0, 0, 0, 0, 0, 0],
                           [0, 0, 0, 0,-1, 1],
                           [0, 0,-1, 1, 0, 0],
                           [0, 0, 1,-1, 0, 0],
                           [0, 0, 0, 0,-1, 1],
                           [0, 0, 0, 0, 0, 0],
                           [-1, 1, 0, 0, 0, 0],
                           [0, 0, 0, 0, 1,-1],
                           [0, 0,-1, 1, 0, 0],
                           [-1, 1, 0, 0, 0, 0],
                           [0, 0, 0, 0, 0, 0]])

### 
UNORIENTED_INTERSECTION_MAT = Matrix([[0, 0, 0],
                                      [1, 0, 0],
                                      [0, 1, 0],
                                      [0, 0, 1],
                                      [1, 0, 0],
                                      [0, 0, 0],
                                      [0, 0, 1],
                                      [0, 1, 0],
                                      [0, 1, 0],
                                      [0, 0, 1],
                                      [0, 0, 0],
                                      [1, 0, 0],
                                      [0, 0, 1],
                                      [0, 1, 0],
                                      [1, 0, 0],
                                      [0, 0, 0]])

ABSNEG_INTERSECTION_MAT = NEG_INTERSECTION_MAT.apply_map(abs)



                  



from . import euler
from . import homology
from . import matrices
from . import boundary
from . import embedded
from . import simplicial
from . import abstract_nbhd
from . import orientable
from . import connected
from . import peripheral
from . import convex_hull






class cached_property(object):
    """
    Descriptor (non-data) for building an attribute on-demand on first use.
    """
    def __init__(self, factory):
        """
        <factory> is called such: factory(instance) to build the attribute.
        """
        self._attr_name = factory.__name__
        self._factory = factory

    def __get__(self, instance, owner):
        # Build the attribute.
        attr = self._factory(instance)

        # Cache the value; hide ourselves.
        setattr(instance, self._attr_name, attr)

        return attr














#!/usr/bin/python

import threading

try:
    import Tkinter as tk
    import tkFileDialog
    import tkMessageBox
except ImportError:  # Python 3.
    try:
        import tkinter as tk
        import tkinter.filedialog as tkFileDialog
        import tkinter.messagebox as tkMessageBox
    except ImportError:
        raise ImportError('Tkinter not available.')

try:
    import ttk as ttk
except ImportError:  # Python 3.
    try:
        from tkinter import ttk
    except ImportError:
        raise ImportError('Ttk not available.')


from tkinter import *

class ConsoleText(tk.Text):
    '''A Tkinter Text widget that provides a scrolling display of console
    stderr and stdout.'''

    class IORedirector(object):
        '''A general class for redirecting I/O to this Text widget.'''
        def __init__(self,text_area):
            self.text_area = text_area

    class StdoutRedirector(IORedirector):
        '''A class for redirecting stdout to this Text widget.'''
        def write(self,str):
            self.text_area.write(str,False)

    class StderrRedirector(IORedirector):
        '''A class for redirecting stderr to this Text widget.'''
        def write(self,str):
            self.text_area.write(str,True)

    def __init__(self, master=None, cnf={}, **kw):
        '''See the __init__ for Tkinter.Text for most of this stuff.'''

        tk.Text.__init__(self, master, cnf, **kw)

        self.started = False
        self.write_lock = threading.Lock()

        self.tag_configure('STDOUT',background='white',foreground='black')
#        self.tag_configure('STDERR',background='white',foreground='red')

        self.config(state=tk.NORMAL)
        self.bind('<Key>',lambda e: 'break') #ignore all key presses

    def start(self):

        if self.started:
            return

        self.started = True

        self.original_stdout = sys.stdout
#        self.original_stderr = sys.stderr

        stdout_redirector = ConsoleText.StdoutRedirector(self)
#        stderr_redirector = ConsoleText.StderrRedirector(self)

        sys.stdout = stdout_redirector
#        sys.stderr = stderr_redirector

    def stop(self):

        if not self.started:
            return

        self.started = False

        sys.stdout = self.original_stdout
#        sys.stderr = self.original_stderr

    def write(self,val,is_stderr=False):

        #Fun Fact:  The way Tkinter Text objects work is that if they're disabled,
        #you can't write into them AT ALL (via the GUI or programatically).  Since we want them
        #disabled for the user, we have to set them to NORMAL (a.k.a. ENABLED), write to them,
        #then set their state back to DISABLED.

        self.write_lock.acquire()

        self.insert('end',val,'STDERR' if is_stderr else 'STDOUT')
        self.see('end')

        self.write_lock.release()


#__________________________________________________________________







from sage.all import Matrix
from sage.all import Integer
from sage.all import vector
from sage.all import Graphics
from sage.all import text3d
from sage.all import Polyhedron
from sage.all import gcd
from sage.all import floor
from sage.all import n
from sage.all import text
from sage.all import RR
from sage.all import VectorSpace
from sage.all import QQ
from sage.all import load_attach_path
from sage.all import pi
from sage.all import sin
from sage.all import cos
from sage.all import arccos
from sage.all import arctan
from sage.all import arcsin
from sage.all import sqrt
from sage.all import span
from sage.all import Graph



from tnorm.utilities.regina_helpers import regina_to_sage_int
from tnorm.kernel.matrices import quads_mat

from tnorm.kernel.abstract_nbhd import abstract_nbhds

from tnorm.utilities.sage_types import Matrix


def orient(spun_normal_surface):
    S = spun_normal_surface
    T = S.triangulation()

    # We'll create a dictionary that keeps track of quad orientations wrt tets. For a given tet, 
    # with k quads, the quads are numbered from 0 to k-1, with the 0th quad being the one that is closest
    # to vertex 0 of tet. Then oriented_quads[tet] is a list of length k whose i^th entry is +1 if the 
    # orientation of the i^th quad agrees with the local orientation ("alignment"), and -1 if it disagrees.
    # At first, all entries are 0 to indicate that orientation is unknown (or really, not yet determined).


    abs_nbhds = abstract_nbhds(S)

    # abs_nbhds stores quads grouped according to the tetrahedron they are in. For below, it will be more
    # useful to store them according to which connected component of the intersection of the surface and 
    # the abstract neighborhood they are in.
    abstract_comps = []
    for e in abs_nbhds:
        N = abs_nbhds[e]
        for comp in N['components']:
            Qs = [i for i in range(len(comp)) if comp[i] != -1]
            abstract_comps.append({'tets':[N['tets'][i] for i in Qs], 'slopes':[N['slopes'][i] for i in Qs], 'align':[N['align'][i] for i in Qs], 'components':[comp[i] for i in Qs]})
    
    # iterator for assigning orientations to quads. See below.
    def iterate(comp, abstract_comps, oriented_quads, orientn_wrt_nbd=1):
        orientable = True

        # we iterate through the normal quadrilateral disks in the component comp, 
        # triangular disks are skips because we have already removed them from comp
        for i in range(len(comp['tets'])):

            # set orientation of this quad with respect to the tet
            orientn_wrt_tet = comp['align'][i]*orientn_wrt_nbd

            # if the orientation of the quad is not already set in "oriented_quads",
            # then set it to orientation found above
            if oriented_quads[comp['tets'][i]][comp['components'][i]] == 0:
                oriented_quads[comp['tets'][i]][comp['components'][i]] = orientn_wrt_tet
            
            # if there is already an orientation on the quad set in "oriented_quads",
            # then make sure it matches the orientation found above. If it doesn't,
            # then set orientalbe to False, and break.
            else:
                if oriented_quads[comp['tets'][i]][comp['components'][i]] != orientn_wrt_tet:
                    orientable = False
                    break

            # for each quad in comp, look for other components that it appears in, and add each 
            # of them, along with the orientation comp induces on them, to the list "new_comps"
            new_comps = []
            for other_comp in abstract_comps:
                for j in range(len(other_comp['tets'])):
                    if comp['tets'][i] == other_comp['tets'][j] and comp['components'][i] == other_comp['components'][j]:
                        new_orientn_wrt_nbd = orientn_wrt_tet*other_comp['align'][j]
                        new_comps.append((other_comp,new_orientn_wrt_nbd))
                        break

            # remove from abstract_comps all the components we just added to new_comps
            for new_comp in new_comps:
                abstract_comps.remove(new_comp[0])

            # for each component in new_comps, run iterate again. If new_comps is empty, then
            # we have exhausted this connected component, so we return
            for new_comp in new_comps:
                abstract_comps, oriented_quads, orientable = iterate(new_comp[0],abstract_comps,oriented_quads,new_comp[1])
                if not orientable:
                    return abstract_comps, oriented_quads, orientable
        return abstract_comps, oriented_quads, orientable

    # pick a first abstract component, and impose a transverse orientation. A transverse orientation is 
    # induced on each quad in the component. For each of these quads, we search for all remaining abstract
    # components in which they appear, and give that component the orientation induced by the orientation
    # on the quad. Iterate until we run out of components, or until we come to a component where the 
    # orientation induced by the quad that brought us there disagrees with the orientation of another quad
    # which has already been oriented (in which cases the surface is not orientable).
    oriented_quads_matrices = []
    while len(abstract_comps) > 0:
        # get a component. When we call iterate() below, we stay in this call to iterate
        # until we get orientations that are not compatible, or the component containing 
        # comp is completely oriented.
        oriented_quads = {}
        for i in range(T.size()):
            num_quads = regina_to_sage_int(S.quads(i,0) + S.quads(i,1) + S.quads(i,2))
            oriented_quads[i] = [0]*num_quads

        comp = abstract_comps.pop()
        abstract_comps, oriented_quads, orientable = iterate(comp, abstract_comps, oriented_quads)
        if not orientable:
            break

        # if orientable is still set to True, then this component is orientable, so we
        # compute the oriented_quads_mat for the component and store it in oriented_quads_matrices.
        if orientable:
            quads = quads_mat(S)
            emb_oriented_quads_mat = []
            for i in range(T.size()):
                row=[]
                for j in [0,1,2]:
                    if quads[i][j] != 0:
                        pos = sum([k for k in oriented_quads[i] if k==1])
                        neg = -sum([k for k in oriented_quads[i] if k==-1])
                        #assert pos+neg == quads[i][j]
                        row.append(pos)
                        row.append(neg)
                    else:
                        row.append(0); row.append(0)
                emb_oriented_quads_mat.append(row)
            oriented_quads_matrices.append(Matrix(emb_oriented_quads_mat))
    if orientable:
        return oriented_quads_matrices
    else:
        return False


def is_orientable(spun_normal_surface):
    S=spun_normal_surface
    if orient(S) != False:
        return True
    else:
        return False



# Example of embedding CEF Python browser using Tkinter toolkit.
# This example has two widgets: a navigation bar and a browser.
#
# NOTE: This example often crashes on Mac (Python 2.7, Tk 8.5/8.6)
#       during initial app loading with such message:
#       "Segmentation fault: 11". Reported as Issue #309.
#
# Tested configurations:
# - Tk 8.5 on Windows/Mac
# - Tk 8.6 on Linux
# - CEF Python v55.3+
#
# Known issue on Linux: When typing url, mouse must be over url
# entry widget otherwise keyboard focus is lost (Issue #255
# and Issue #284).

from cefpython3 import cefpython as cef
import ctypes
try:
    import tkinter as tk
except ImportError:
    import Tkinter as tk
import sys
import os
import platform
import logging as _logging

# Fix for PyCharm hints warnings
WindowUtils = cef.WindowUtils()

# Platforms
WINDOWS = (platform.system() == "Windows")
LINUX = (platform.system() == "Linux")
MAC = (platform.system() == "Darwin")

# Globals
logger = _logging.getLogger("tkinter_.py")

# Constants
# Tk 8.5 doesn't support png images
IMAGE_EXT = ".png" if tk.TkVersion > 8.5 else ".gif"


class MainFrame(tk.Frame):

    def __init__(self, root):
        self.browser_frame = None
        self.navigation_bar = None

        # Root
        root.geometry("900x640")
        tk.Grid.rowconfigure(root, 0, weight=1)
        tk.Grid.columnconfigure(root, 0, weight=1)

        # MainFrame
        tk.Frame.__init__(self, root)
        self.master.title("Tkinter example")
        self.master.protocol("WM_DELETE_WINDOW", self.on_close)
        self.master.bind("<Configure>", self.on_root_configure)
        self.setup_icon()
        self.bind("<Configure>", self.on_configure)
        self.bind("<FocusIn>", self.on_focus_in)
        self.bind("<FocusOut>", self.on_focus_out)

        # NavigationBar
        self.navigation_bar = NavigationBar(self)
        self.navigation_bar.grid(row=0, column=0,
                                 sticky=(tk.N + tk.S + tk.E + tk.W))
        tk.Grid.rowconfigure(self, 0, weight=0)
        tk.Grid.columnconfigure(self, 0, weight=0)

        # BrowserFrame
        self.browser_frame = BrowserFrame(self, self.navigation_bar)
        self.browser_frame.grid(row=1, column=0,
                                sticky=(tk.N + tk.S + tk.E + tk.W))
        tk.Grid.rowconfigure(self, 1, weight=1)
        tk.Grid.columnconfigure(self, 0, weight=1)

        # Pack MainFrame
        self.pack(fill=tk.BOTH, expand=tk.YES)

    def on_root_configure(self, _):
        logger.debug("MainFrame.on_root_configure")
        if self.browser_frame:
            self.browser_frame.on_root_configure()

    def on_configure(self, event):
        logger.debug("MainFrame.on_configure")
        if self.browser_frame:
            width = event.width
            height = event.height
            if self.navigation_bar:
                height = height - self.navigation_bar.winfo_height()
            self.browser_frame.on_mainframe_configure(width, height)

    def on_focus_in(self, _):
        logger.debug("MainFrame.on_focus_in")

    def on_focus_out(self, _):
        logger.debug("MainFrame.on_focus_out")

    def on_close(self):
        if self.browser_frame:
            self.browser_frame.on_root_close()
        self.master.destroy()

    def get_browser(self):
        if self.browser_frame:
            return self.browser_frame.browser
        return None

    def get_browser_frame(self):
        if self.browser_frame:
            return self.browser_frame
        return None

    def setup_icon(self):
        resources = os.path.join(os.path.dirname(__file__), "resources")
        icon_path = os.path.join(resources, "tkinter"+IMAGE_EXT)
        if os.path.exists(icon_path):
            self.icon = tk.PhotoImage(file=icon_path)
            # noinspection PyProtectedMember
            self.master.call("wm", "iconphoto", self.master._w, self.icon)


class BrowserFrame(tk.Frame):

    def __init__(self, master, navigation_bar=None):
        self.navigation_bar = navigation_bar
        self.closing = False
        self.browser = None
        tk.Frame.__init__(self, master)
        self.bind("<FocusIn>", self.on_focus_in)
        self.bind("<FocusOut>", self.on_focus_out)
        self.bind("<Configure>", self.on_configure)
        self.focus_set()

    def embed_browser(self):
        window_info = cef.WindowInfo()
        rect = [0, 0, self.winfo_width(), self.winfo_height()]
        window_info.SetAsChild(self.get_window_handle(), rect)
        self.browser = cef.CreateBrowserSync(window_info,
                                             url="file:///J:\q.htm") #todo
        assert self.browser
        self.browser.SetClientHandler(LoadHandler(self))
        self.browser.SetClientHandler(FocusHandler(self))
        self.message_loop_work()

    def get_window_handle(self):
        if self.winfo_id() > 0:
            return self.winfo_id()
        elif MAC:
            # On Mac window id is an invalid negative value (Issue #308).
            # This is kind of a dirty hack to get window handle using
            # PyObjC package. If you change structure of windows then you
            # need to do modifications here as well.
            # noinspection PyUnresolvedReferences
            from AppKit import NSApp
            # noinspection PyUnresolvedReferences
            import objc
            # Sometimes there is more than one window, when application
            # didn't close cleanly last time Python displays an NSAlert
            # window asking whether to Reopen that window.
            # noinspection PyUnresolvedReferences
            return objc.pyobjc_id(NSApp.windows()[-1].contentView())
        else:
            raise Exception("Couldn't obtain window handle")

    def message_loop_work(self):
        cef.MessageLoopWork()
        self.after(10, self.message_loop_work)

    def on_configure(self, _):
        if not self.browser:
            self.embed_browser()

    def on_root_configure(self):
        # Root <Configure> event will be called when top window is moved
        if self.browser:
            self.browser.NotifyMoveOrResizeStarted()

    def on_mainframe_configure(self, width, height):
        if self.browser:
            if WINDOWS:
                ctypes.windll.user32.SetWindowPos(
                    self.browser.GetWindowHandle(), 0,
                    0, 0, width, height, 0x0002)
            elif LINUX:
                self.browser.SetBounds(0, 0, width, height)
            self.browser.NotifyMoveOrResizeStarted()

    def on_focus_in(self, _):
        logger.debug("BrowserFrame.on_focus_in")
        if self.browser:
            self.browser.SetFocus(True)

    def on_focus_out(self, _):
        logger.debug("BrowserFrame.on_focus_out")
        if self.browser:
            self.browser.SetFocus(False)

    def on_root_close(self):
        if self.browser:
            self.browser.CloseBrowser(True)
            self.clear_browser_references()
        self.destroy()

    def clear_browser_references(self):
        # Clear browser references that you keep anywhere in your
        # code. All references must be cleared for CEF to shutdown cleanly.
        self.browser = None


class LoadHandler(object):

    def __init__(self, browser_frame):
        self.browser_frame = browser_frame

    def OnLoadStart(self, browser, **_):
        if self.browser_frame.master.navigation_bar:
            self.browser_frame.master.navigation_bar.set_url(browser.GetUrl())


class FocusHandler(object):

    def __init__(self, browser_frame):
        self.browser_frame = browser_frame

    def OnTakeFocus(self, next_component, **_):
        logger.debug("FocusHandler.OnTakeFocus, next={next}"
                     .format(next=next_component))

    def OnSetFocus(self, source, **_):
        logger.debug("FocusHandler.OnSetFocus, source={source}"
                     .format(source=source))
        return False

    def OnGotFocus(self, **_):
        """Fix CEF focus issues (#255). Call browser frame's focus_set
           to get rid of type cursor in url entry widget."""
        logger.debug("FocusHandler.OnGotFocus")
        self.browser_frame.focus_set()


class NavigationBar(tk.Frame):
    def __init__(self, master):
        self.back_state = tk.NONE
        self.forward_state = tk.NONE
        self.back_image = None
        self.forward_image = None
        self.reload_image = None

        tk.Frame.__init__(self, master)
        resources = os.path.join(os.path.dirname(__file__), "resources")

        # Back button
        back_png = os.path.join(resources, "back"+IMAGE_EXT)
        if os.path.exists(back_png):
            self.back_image = tk.PhotoImage(file=back_png)
        self.back_button = tk.Button(self, image=self.back_image,
                                     command=self.go_back)
        self.back_button.grid(row=0, column=0)

        # Forward button
        forward_png = os.path.join(resources, "forward"+IMAGE_EXT)
        if os.path.exists(forward_png):
            self.forward_image = tk.PhotoImage(file=forward_png)
        self.forward_button = tk.Button(self, image=self.forward_image,
                                        command=self.go_forward)
        self.forward_button.grid(row=0, column=1)

        # Reload button
        reload_png = os.path.join(resources, "reload"+IMAGE_EXT)
        if os.path.exists(reload_png):
            self.reload_image = tk.PhotoImage(file=reload_png)
        self.reload_button = tk.Button(self, image=self.reload_image,
                                       command=self.reload)
        self.reload_button.grid(row=0, column=2)

        # Url entry
        self.url_entry = tk.Entry(self)
        self.url_entry.bind("<FocusIn>", self.on_url_focus_in)
        self.url_entry.bind("<FocusOut>", self.on_url_focus_out)
        self.url_entry.bind("<Return>", self.on_load_url)
        self.url_entry.bind("<Button-1>", self.on_button1)
        self.url_entry.grid(row=0, column=3,
                            sticky=(tk.N + tk.S + tk.E + tk.W))
        tk.Grid.rowconfigure(self, 0, weight=100)
        tk.Grid.columnconfigure(self, 3, weight=100)

        # Update state of buttons
        self.update_state()

    def go_back(self):
        if self.master.get_browser():
            self.master.get_browser().GoBack()

    def go_forward(self):
        if self.master.get_browser():
            self.master.get_browser().GoForward()

    def reload(self):
        if self.master.get_browser():
            self.master.get_browser().Reload()

    def set_url(self, url):
        self.url_entry.delete(0, tk.END)
        self.url_entry.insert(0, url)

    def on_url_focus_in(self, _):
        logger.debug("NavigationBar.on_url_focus_in")

    def on_url_focus_out(self, _):
        logger.debug("NavigationBar.on_url_focus_out")

    def on_load_url(self, _):
        if self.master.get_browser():
            self.master.get_browser().StopLoad()
            self.master.get_browser().LoadUrl(self.url_entry.get())

    def on_button1(self, _):
        """Fix CEF focus issues (#255). See also FocusHandler.OnGotFocus."""
        logger.debug("NavigationBar.on_button1")
        self.master.master.focus_force()

    def update_state(self):
        browser = self.master.get_browser()
        if not browser:
            if self.back_state != tk.DISABLED:
                self.back_button.config(state=tk.DISABLED)
                self.back_state = tk.DISABLED
            if self.forward_state != tk.DISABLED:
                self.forward_button.config(state=tk.DISABLED)
                self.forward_state = tk.DISABLED
            self.after(100, self.update_state)
            return
        if browser.CanGoBack():
            if self.back_state != tk.NORMAL:
                self.back_button.config(state=tk.NORMAL)
                self.back_state = tk.NORMAL
        else:
            if self.back_state != tk.DISABLED:
                self.back_button.config(state=tk.DISABLED)
                self.back_state = tk.DISABLED
        if browser.CanGoForward():
            if self.forward_state != tk.NORMAL:
                self.forward_button.config(state=tk.NORMAL)
                self.forward_state = tk.NORMAL
        else:
            if self.forward_state != tk.DISABLED:
                self.forward_button.config(state=tk.DISABLED)
                self.forward_state = tk.DISABLED
        self.after(100, self.update_state)


if __name__ == '__main__':
    logger.setLevel(_logging.INFO)
    stream_handler = _logging.StreamHandler()
    formatter = _logging.Formatter("[%(filename)s] %(message)s")
    stream_handler.setFormatter(formatter)
    logger.addHandler(stream_handler)
    logger.info("CEF Python {ver}".format(ver=cef.__version__))
    logger.info("Python {ver} {arch}".format(
            ver=platform.python_version(), arch=platform.architecture()[0]))
    logger.info("Tk {ver}".format(ver=tk.Tcl().eval('info patchlevel')))
    assert cef.__version__ >= "55.3", "CEF Python v55.3+ required to run this"
    sys.excepthook = cef.ExceptHook  # To shutdown all CEF processes on error
    root = tk.Tk()
    app = MainFrame(root)
    # Tk must be initialized before CEF otherwise fatal error (Issue #306)
    cef.Initialize()

    app.mainloop()
    cef.Shutdown()



from tnorm.utilities.sage_types import vector


def _map_to_H1bdy(oriented_spun_surface, TN_wrapper):
    s = oriented_spun_surface
    pos_slopes, neg_slopes = TN_wrapper.boundary_slopes(s).values()

    slopes = []
    for i in range(len(pos_slopes)):
        slope = (pos_slopes[i][0]+neg_slopes[i][0], pos_slopes[i][1]+neg_slopes[i][1])
        slopes.append(slope)

    return slopes


### this maps an oriented spun normal surface to H_2(M,\partial M; \ZZ)
def _map_to_H2(oriented_spun_surface, TN_wrapper):
    s = oriented_spun_surface
    slopes = TN_wrapper.H1bdy_slopes(s)
    coords = vector([slopes[i][1] for i in range(len(slopes))])
    return coords


#################--------------------------------------------------------------------






from math import gcd as GCD
import itertools

from tnorm.kernel import embedded, orientable, connected, euler
from tnorm.kernel.matrices import peripheral_curve_mats, oriented_quads_mat, quads_mat
from tnorm.kernel.boundary import inward_oriented_bdy, outward_oriented_bdy, unoriented_spinning_slopes
import snappy

from tnorm.utilities.sage_types import Matrix

from tnorm.utilities.regina_helpers import regina_to_sage_int


def orient(S):
    return orientable.orient(S)

def is_orientable(S):
    return orientable.is_orientable(S)

def euler_char(S):
    T = S.triangulation()
    assert T.isIdeal(), 'the triangulation must be ideal.'
    return euler.euler_char_(S,euler.solve_lin_gluing_eq(T))

def spinning_slopes(S):
    q_mat = quads_mat(S)
    T = S.triangulation()
    M = snappy.Manifold(T.snapPea())
    pc_mats = peripheral_curve_mats(M,T)
    slopes = unoriented_spinning_slopes(S, pc_mats, q_mat)
    return slopes

def num_boundary_comps(S):
    slopes = spinning_slopes(S)
    return sum([gcd(slope[0],slope[1]) for slope in slopes])

def oriented_boundary_slopes(S, orientation=None):
    if orientation == None:
        oriented_quads_mat = sum(orient(S))
    else:
        oriented_quads_mat = orientation
    assert oriented_quads_mat != False, 'the surface is not orientable.'
    T = S.triangulation()
    M = snappy.Manifold(T.snapPea())
    pc_mats = peripheral_curve_mats(M,T)
    out_bdy = outward_oriented_bdy(S, pc_mats, oriented_quads_mat, False)
    in_bdy = inward_oriented_bdy(S, pc_mats, oriented_quads_mat, False)

    return {'outward':out_bdy, 'inward':in_bdy}, oriented_quads_mat

def is_connected(S):
    return connected.is_connected(S)

def connected_components(S):
    return connected.connected_components(S)

def haken_sum(R,S):
    if R.locallyCompatible(S):
        return connected.sum(R,S)
    else:
        return "these surfaces are not locally compatible!"


def gcd(a,b):
    return abs(GCD(a,b))


from math import gcd as GCD
import itertools

from tnorm.utilities.sage_types import Matrix
from tnorm.kernel.matrices import oriented_quads_mat
from tnorm.kernel.orientable import orient



def ends_embedded(qtons, TN_wrapper):
    outward_bdy, inward_bdy = TN_wrapper.boundary_slopes(qtons).values()
    for i in range(len(outward_bdy)):
        gcd_outward = gcd(outward_bdy[i][0],outward_bdy[i][1])
        gcd_inward = gcd(inward_bdy[i][0],inward_bdy[i][1])
        if gcd_outward != 0 and gcd_inward != 0:
            if not ( outward_bdy[i][0]/gcd_outward == - inward_bdy[i][0]/gcd_inward and outward_bdy[i][1]/gcd_outward == - inward_bdy[i][1]/gcd_inward ):
                return False
    return True

def is_embedded(qtons,TN_wrapper):
    W = TN_wrapper
    S = qtons
    T = W.triangulation()

    if not W.manifold_is_closed():
        if not ends_embedded(S,W):
            return False

    oriented_projection = orient(S)

    if oriented_projection != False:
        k = len(oriented_projection)
        q_mat = oriented_quads_mat(S)

        orientations = list(itertools.product([1,-1], repeat=k))
        for nu in orientations:
            nu_oriented_proj = Matrix([[0 for i in range(6)] for j in range(T.size())])
            for i in range(k):
                if nu[i] == -1:
                    comp_orientation = opp_oriented_quads_mat(oriented_projection[i])
                else:
                    comp_orientation = oriented_projection[i]
                nu_oriented_proj += comp_orientation

            if nu_oriented_proj == q_mat:
                return True
  
    return False

def opp_oriented_quads_mat(oriented_quads_mat):
    opp_oriented_quads_mat = []
    for row in oriented_quads_mat:
        opp_row = []
        for i in [0,2,4]:
            opp_row.append(row[i+1])
            opp_row.append(row[i])
        opp_oriented_quads_mat.append(opp_row)
    return Matrix(opp_oriented_quads_mat)


def gcd(a,b):
    return abs(GCD(a,b))

















from tnorm.kernel.abstract_nbhd import abstract_nbhds
from tnorm.kernel.matrices import quads_mat


from tnorm.utilities.sage_types import Graph

import regina


def connected_components(spun_normal_surface):
    S = spun_normal_surface
    T = S.triangulation()

    abst_nbhds = abstract_nbhds(S)
    quads_matrix = quads_mat(S)
    abst_nbhd_comps = [set([(abst_nbhds[k]['tets'][j],comp[j]) for j in range(len(comp)) if comp[j]!=-1]) for k in abst_nbhds for comp in abst_nbhds[k]['components']]

    anc_dict = dict([(i,abst_nbhd_comps[i]) for i in range(len(abst_nbhd_comps))])

    G = Graph(len(anc_dict))

    for i in G.vertices():
        for j in range(i+1, G.num_verts()):
            if len(anc_dict[i].intersection(anc_dict[j])) > 0:
                G.add_edge((i,j))

    components = []
    for C in G.connected_components():
        normal_coords = [int(0) for i in range(3*T.size())]
        quads = anc_dict[C[0]]
        for i in range(1,len(C)):
            quads = quads.union(anc_dict[C[i]])
        for (k,j) in quads:
            for i in range(3):
                if quads_matrix[k][i] != 0:
                    normal_coords[3*k+i] += 1
        F = regina.NormalSurface(T,regina.NS_QUAD, normal_coords)
        components.append(F)

    return components

def is_connected(spun_normal_surface):
    S = spun_normal_surface
    T = S.triangulation()

    abst_nbhds = abstract_nbhds(S)
    quads_matrix = quads_mat(S)
    abst_nbhd_comps = [set([(abst_nbhds[k]['tets'][j],comp[j]) for j in range(len(comp)) if comp[j]!=-1]) for k in abst_nbhds for comp in abst_nbhds[k]['components']]

    anc_dict = dict([(i,abst_nbhd_comps[i]) for i in range(len(abst_nbhd_comps))])

    G = Graph(len(anc_dict))

    for i in G.vertices():
        for j in range(i+1, G.num_verts()):
            if len(anc_dict[i].intersection(anc_dict[j])) > 0:
                G.add_edge((i,j))

    if G.is_connected():
        return True

    else:
        return False

def sum(R, S):

    T = R.triangulation()
    R_quads = quads_mat(R)
    S_quads = quads_mat(S)

    F_quads = R_quads + S_quads

    normal_coords = [int(F_quads[i][j]) for i in range(F_quads.dimensions()[0]) for j in range(3)]

    F = regina.NormalSurface(T, regina.NS_QUAD, normal_coords)

    return F




from tnorm.utilities.regina_helpers import regina_to_sage_int


from tnorm.utilities.sage_types import Matrix



def abstract_nbhds(spun_normal_surface):
    S = spun_normal_surface
    T = S.triangulation()

    quad_counts = {}
    tets_around_edges = {}

    # First we'll gather, for each edge, information about the quads in the tetrahedra around the edge.
    # The tetrahedra are cylically ordered counter-clockwise w.r.t. the orientation of the edge with 
    # vertex on top. 
    for edge in T.faces(int(1)):

        # for each tet around edge, how many quads and what is slope (sign).
        quad_count = []

        # for each tet around edge, a tuple (tet_index, edge_type, edge embedding permutation)
        tets_around_edge = []

        embs = cyclically_order_embeddings(edge.embeddings(),T)
        for f in embs:
            t = f.tetrahedron()
            e = f.edge()
            p = f.vertices()
            tets_around_edge.append((t.index(),e,p))
            types = {0:[1,2], 5:[1,2], 1:[2,0], 4:[2,0], 2:[0,1], 3:[0,1]}
            for i in types[e]:
                count = regina_to_sage_int(S.quads(t.index(),i))
                if count != 0:
                    if i == types[e][1]:
                        count = -count
                    break
            quad_count.append(count)
        quad_counts[edge.index()] = quad_count
        tets_around_edges[edge.index()] = tets_around_edge

    # width is the number of connected components of the surface in the abstract nbhd at the edge,
    # ind is the index in the cyclic ordering around the edge where adding up crossing counts
    # from i to some later j gives the width. 
    widths = {}
    for edge in quad_counts:
        width, ind = 0, 0
        L = len(quad_counts[edge])
        for i in range(L):
            for j in range(i,L):
                this_width = int(sum(quad_counts[edge][i:j+1]))
                if abs(this_width) > width:
                    if this_width < 0:
                        width, ind = abs(this_width), j+1
                    else:
                        width, ind = this_width, i
        widths[edge] = (ind, width)

    # for each edge, abst_nbhds stores the info we need about the quads in its abstract nbhd.
    abst_nbhds = {}
    for edge in T.faces(int(1)):

        # for this edge, info we want to store about the abstract nbhd
        abstract_nbhd = []

        # get the counts for this edge, computed above
        counts = quad_counts[edge.index()]
        L = len(counts)

        # get ind and width for this edge, computed above
        ind, width = widths[edge.index()]

        # get tets_around_edge for this edge, computed above
        tets = tets_around_edges[edge.index()]

        # if there is anything in this abstract nbhd, continue
        if width > 0:

            # if there are no quads in the tet at ind in the ordering, then increasing ind until there are
            while counts[ind] == 0:
                ind = (ind + 1) % L
            assert counts[ind] > 0

            # For the first tet cyclically forward from ind containing quads, the quads will have positive
            # slope (because of how ind was chosen above), and will be as far toward vertex 1 of the edge as 
            # possible (i.e., in the top connected  component of the intersection of surface with abstract nbhd.)
            slope = 1
                
            tet, edge_type, perm = tets[ind]

            # For a permutation perm that described how edge is embedded in tet (or, equivalently, how tet 
            # is embedded in the abstract nbhd), this function tells us if the 0 vertex of tet is below the quad
            # (in which case alignment(perm)=-1) or above the quad (in which case alignment(perm)=1). This can
            # be interpreted as a transerve orientation of the quad, which gives a frame of reference. When we 
            # orient the quads, we can describe the orientation based on whether or not it agrees with this 
            # orientation.
            def alignment(perm):
                if perm[0]==0 or (slope==-1 and perm[3]==0) or (slope==1 and perm[2]==0):
                    return 1
                else:
                    return -1

            align = alignment(perm)

            # arranged quads is a vector of length width, whose components are in range(counts[ind]) for a quad, and -1 for a triangle.
            # e.g., vector [-1,0,1,-1] means in the tet at ind in cyclic ordering, top layer is a triangle, next down is
            # the 0th quad, next is 1st quad, next is a triangle (as we go along the edge from vertex 1 to vertex 0)
            arranged_quads = quad_range(counts[ind],align) + [-1]*(width-counts[ind])

            # where in the vertical ordering of the surface components are the first and last quad.
            first_quad, last_quad = 0,counts[ind]-1

            abstract_nbhd.append((tet,slope,align,arranged_quads,perm))

                
            # now continue, in the cyclic ordering, until we go all the way around the edge.
            while len(abstract_nbhd) < L:
                prev_slope = slope
                if counts[ind] != 0:
                    prev_count = counts[ind]
                ind = (ind + 1) % L

                if counts[ind] > 0:
                    slope = 1
                elif counts[ind] < 0:
                    slope = -1
                else: 
                    slope = prev_slope

                tet, edge_type, perm = tets[ind]
                align = alignment(perm)
                if counts[ind] == 0:
                    arranged_quads = [-1]*width
                else:
                    if prev_slope == 1 and slope == 1:
                        arranged_quads = pad_with_tris([-1]*(last_quad+1) + quad_range(counts[ind],align), width)
                    elif prev_slope == 1 and slope == -1:
                        arranged_quads = pad_with_tris([-1]*((prev_count + counts[ind]) + first_quad) + quad_range(counts[ind],align), width)
                    elif prev_slope == -1 and slope == 1:
                        arranged_quads = pad_with_tris([-1]*(first_quad) + quad_range(counts[ind],align), width)
                    elif prev_slope == -1 and slope == -1:
                        arranged_quads = pad_with_tris([-1]*(first_quad + counts[ind]) + quad_range(counts[ind],align), width)
                    first_quad, last_quad = first_non_neg(arranged_quads), last_non_neg(arranged_quads)
                abstract_nbhd.append((tet,slope,align,arranged_quads,perm))

        N = abstract_nbhd
        tets = [N[i][0] for i in range(len(N))]
        perms = [N[i][4] for i in range(len(N))]
        slopes =  [N[i][1] for i in range(len(N))]
        align = [N[i][2] for i in range(len(N))]
        components = Matrix([[N[i][3][j] for i in range(len(N))] for j in range(width)])

        # check to make sure in each component the quads alternate between positive and negative slopes. If
        # they don't then there is a problem, as in that case matching equations would not hold.
        for comp in components:
            total_sign=0
            for i in range(len(comp)):
                if comp[i] != -1:
                    total_sign += slopes[i]
                    assert total_sign in [-1,0,1], 'something is wrong---please let the developer know about this!'
            assert total_sign == 0, 'something is wrong---please let the developer know about this!'

            
        abst_nbhds[edge.index()] = {'perms':perms,'tets':tets,'slopes':slopes,'align':align,'components':components}

    return abst_nbhds








def quad_range(length,reversed):
    L = [i for i in range(abs(length))]
    if reversed == 1:
        L.reverse()
    return L


def first_non_neg(lst):
    i=0
    while lst[i] < 0 and i < len(lst)-1:
        i += 1
    if lst[i] >= 0:
        return i
    else:
        return None

def last_non_neg(lst):
    i = len(lst)-1
    while lst[i] < 0 and i > 0:
        i -= 1
    if lst[i] >= 0:
        return i
    else:
        return None



def pad_with_tris(lst, total_len):
    N = total_len - len(lst)
    return lst+[-1]*N

def cyclically_order_embeddings(edge_embeddings, T):
    """Order the edge embeddings cyclically
    """
    embs = edge_embeddings
    f = embs[0]
    e = f.edge()
    t = f.tetrahedron()
    dec = f.vertices()[0] < f.vertices()[1]
    ordered_embeddings = [f]
    emb_dict = dict([((f0.tetrahedron().index(),f0.edge(), f0.vertices()[0]<f0.vertices()[1]),f0) for f0 in embs])
    edge_dict = {(0,1):0, (0,2):1, (0,3):2, (1,2):3, (1,3):4, (2,3):5}
    gluing_dict = {(0,True):3, (0,False):2, (1,True):1, (1,False):3, (2,True):2, (2,False):1, (3,True):3, (3,False):0, (4,True):0, (4,False):2, (5,True):1, (5,False):0}
    while len(ordered_embeddings) < len(embs):
        gluing = t.adjacentGluing(gluing_dict[e,dec])
        t = t.adjacentTetrahedron(gluing_dict[e,dec])
        try:
            e = edge_dict[(gluing[f.vertices()[0]],gluing[f.vertices()[1]])]
            dec = True
        except KeyError:
            e = edge_dict[(gluing[f.vertices()[1]],gluing[f.vertices()[0]])]
            dec = False
        f = emb_dict[(t.index(),e,dec)]
        ordered_embeddings.append(f)
    return ordered_embeddings




from __future__ import print_function
from math import ceil

import tkinter as tk
from tkinter import ttk




def make_qtons_table(tnorm_app):
#    try:
    qd = tnorm_app.wrapper.qtons_info
    treedata = [(i, qd[i]['image_in_H2'], qd[i]['euler_char'], 'S_{},{}'.format(qd[i]['genus'], qd[i]['num_boundary_comps']), qd[i]['is_embedded'], str(qd[i]['boundary_slopes']['outward']), str(qd[i]['boundary_slopes']['inward']), qd[i]['is_norm_minimizing'], qd[i]['over_facet']) for i in qd]
    column_names = ('qtons index', 'coords', 'euler char', 'topology', 'is_embedded', 'outward_bdy', 'inward_bdy','norm minimizing?','over facet')
    make_treeview_table(tnorm_app.QtonsTab,treedata,column_names, True)
#    except Exception as e: print('Error: {}'.format(e))
    tnorm_app.stop_spin()


def make_all_vertices_table(tnorm_app, frame):
    #try:
    make_vertices_table(tnorm_app, frame, 'all', True)
    #except Exception as e: print('Error: {}'.format(e))
    tnorm_app.stop_spin()

def make_vertices_table(tnorm_app, frame, vert_list, with_scrollbar=True):
    W = tnorm_app.wrapper
    B = tnorm_app.ball
    if vert_list == 'all':
        vert_list = range(len(B.vertices()))
    height = len(vert_list)
    verts = {i:B.vertices()[i] for i in vert_list}
    column_names = ('vertex', 'qtons index','coords','topology','euler char', 'is_embedded','outward_bdy','inward_bdy')
    treedata = [(i, verts[i].qtons_index(), verts[i].coords(), 'S_{},{}'.format(verts[i].genus(), verts[i].num_boundary_comps()), verts[i].euler_char(), verts[i].is_embedded(), str(verts[i].boundary_slopes()['outward']), str(verts[i].boundary_slopes()['inward'])) for i in vert_list]
    make_treeview_table(frame, treedata, column_names, with_scrollbar)

def make_facets_table(tnorm_app, frame, dim):
    if dim in [0,tnorm_app.ball.dimension()]:
        make_vertices_table(tnorm_app, frame, 'all')
    else:
        W = tnorm_app.wrapper
        B = tnorm_app.ball
        facets = B.facets(dim)
        column_names = ('facet','vertex','qtons index','coords','topology','euler char','is_embedded', 'outward_bdy','inward_bdy')
        width = len(column_names)
        treedata = []
        for j in range(len(facets)):
            f = facets[j]
            vert_list = [v.index() for v in f.vertices()]
            verts = {i:B.vertices()[i] for i in vert_list}
            v0 = vert_list[0]
            treedata.append((j,v0, verts[v0].qtons_index(), verts[v0].coords(), 'S_{},{}'.format(verts[v0].genus(), verts[v0].num_boundary_comps()), verts[v0].euler_char(), verts[v0].is_embedded(), str(verts[v0].boundary_slopes()['outward']), str(verts[v0].boundary_slopes()['inward'])))
            for i in vert_list[1:]:
                treedata.append(('',i, verts[i].qtons_index(), verts[i].coords(), 'S_{},{}'.format(verts[i].genus(), verts[i].num_boundary_comps()), verts[i].euler_char(), verts[i].is_embedded(), str(verts[i].boundary_slopes()['outward']), str(verts[i].boundary_slopes()['inward'])))
            treedata.append(tuple('' for i in range(width)))
        make_treeview_table(frame, treedata, column_names)
    tnorm_app.stop_spin()

def make_dnb_vertices_table(tnorm_app, frame):
    W = tnorm_app.wrapper
    B = tnorm_app.ball
    DB = tnorm_app.dual_ball
    dim = B.dimension() - 1
    facets = B.facets(dim)
    assert DB.num_vertices() == len(facets)
    column_names = ('vertex', 'dual to face','vertices of dual face','qtons_index','coords', 'topology','euler char', 'outward_bdy', 'inward_bdy')
    width = len(column_names)
    treedata = []
    for j in range(len(facets)):
        f = facets[j]
        if dim == 0:
            vert_list = [f.index()]
        else:
            vert_list = [v.index() for v in f.vertices()]
        verts = {i:B.vertices()[i] for i in vert_list}
        v0 = vert_list[0]
        treedata.append((j,j,v0, verts[v0].qtons_index(), verts[v0].coords(), 'S_{},{}'.format(verts[v0].genus(), verts[v0].num_boundary_comps()), verts[v0].euler_char(), str(verts[v0].boundary_slopes()['outward']), str(verts[v0].boundary_slopes()['inward'])))
        for i in vert_list[1:]:
            treedata.append(('','',i, verts[i].qtons_index(), verts[i].coords(), 'S_{},{}'.format(verts[i].genus(), verts[i].num_boundary_comps()), verts[i].euler_char(), str(verts[i].boundary_slopes()['outward']), str(verts[v0].boundary_slopes()['inward'])))
        treedata.append(tuple('' for i in range(width)))
    make_treeview_table(frame, treedata, column_names)
    tnorm_app.stop_spin()


def make_treeview_table(frame, treedata, column_names, with_scrollbar=True):
    if with_scrollbar:
        xscrollbar = ttk.Scrollbar(frame, orient='horizontal')
        yscrollbar = ttk.Scrollbar(frame, orient='vertical')
        tree = ttk.Treeview(frame, columns=column_names, yscrollcommand=yscrollbar.set, xscrollcommand=xscrollbar.set, show=['headings'])
    else:
        tree = ttk.Treeview(frame, columns=column_names, show=['headings'])
    for i in range(len(column_names)):
        col = column_names[i]
        width = int(ceil(max([len(col), max([len(str(treedata[j][i])) for j in range(len(treedata))])])*7))
        tree.heading(col, text=col, anchor='center')
        tree.column(col, minwidth=0, width=width, stretch=tk.NO, anchor='center')
    for x in treedata:
        tree.insert('', 'end', values=x)
    frame.columnconfigure(0, weight=1) # column with treeview
    frame.rowconfigure(0, weight=1) # row with treeview  
    tree.grid(row=0, column=0, sticky='nswe')
    if with_scrollbar:
        xscrollbar.config(command=tree.xview)
        yscrollbar.config(command=tree.yview)
        yscrollbar.grid(row=0, column=1, sticky='ns')
        xscrollbar.grid(row=1, column=0, sticky='we')







import snappy
import tnorm.kernel.simplicial
import regina
import random


has_internal_homology = ['s789', 'v1539', 'v3209', 't06828', 't07788', 't08868', 't09692', 't12061','o9_09412', 'o9_23005', 'o9_23006', 'o9_31901', 'o9_32877', 'o9_36416', 'o9_36417', 'o9_36580', 'o9_40732', 'o9_41691', 'o9_41693', 'o9_44004', 'o9_44007', 'o9_44157']

closed_isosigs = ['kLLzvQQkacegjihihjjnkxmeqxjjcn', 'kLvPPPAkafedhhiigjjjqxnqxahnhj', 'kLvzQQzkccfehghgijjhnhqckcnahn', 'kLvzQQzkccfehghgijjhrhqckcnahr', 'kLLLAPPkcdgfehijijjhshaqpihkkk', 'kLLwvQQkacdgijhjiijjkalilmckwo', 'kLLzLPQkacegjihihjjjkxpuqxrwvv', 'kLLzLPQkacegjihihjjnkxpuqxrwvv', 'kLLzLPQkaceihjgijjijkxuetujjgs', 'kLLzLPQkaceihjgijjinkxuetujjgs']

links = ['L14n43208', 'L12n2026', 'L12n1054', 'L14n49678', 'L7a6', 'L8a18', 'L9a46']

  
nonfibered = ['K5a1','K6a3','K7a4','K7a5','K7a6','K7a3','K8a11','K8a18','K8a17','K8a10','K8a4','K8a9','K8a7','K8a1','K8a2','K9a27','K9a38','K9a35','K9a36','K9a23','K9a26','K9a8','K9a33','K9a39','K9a22','K9a34','K9a17','K9a10','K9a25','K9a24','K9a3','K9a21','K9a16','K9a4','K9a40','K9a18','K9a30','K9a32','K9a29','K9n5','K9n8','K10a75','K10a117','K10a113','K10a70','K10a65','K10a114','K10a64','K10a116','K10a43','K10a54','K10a33','K10a68','K10a115','K10a63','K10a108','K10a74','K10a60','K10a112','K10a57','K10a71','K10a61','K10a111','K10a58','K10a44','K10a34','K10a69','K10a55','K10a109','K10a19','K10a23','K10a5','K10a49','K10a29','K10a26','K10a30','K10a13','K10a82','K10a16','K10a80','K11a1','K11a2','K11a4','K11a6','K11a8','K11a10','K11a11','K11a12','K11a13','K11a16','K11a18','K11a20','K11a21','K11a23','K11a27','K11a29','K11a30','K11a31','K11a32','K11a36','K11a37','K11a38','K11a39','K11a41','K11a43','K11a45','K11a46','K11a48','K11a49','K11a50','K11a52','K11a54','K11a56','K11a58','K11a59','K11a60']


fibered = ['K3a1','K4a1','K5a2','K6a2','K6a1','K7a7','K7a2','K7a1','K8a8','K8a13','K8a6','K8a16','K8a3','K8a5','K8a15','K8a14','K8a12','K8n3','K8n1','K8n2','K9a41','K9a20','K9a14','K9a19','K9a2','K9a7','K9a15','K9a12','K9a5','K9a31','K9a1','K9a13','K9a6','K9a11','K9a28','K9a9','K9a37','K9n4','K9n3','K9n1','K9n2','K9n7','K9n6','K10a59','K10a56','K10a110','K10a107','K10a53','K10a35','K10a31','K10a52','K10a32','K10a25','K10a81','K10a15','K10a79','K10a2','K10a1','K10a41','K10a122','K10a38','K10a22','K10a10','K10a3','K10a27','K10a17','K10a78','K10a7','K10a83','K10a86','K10a11','K10a21','K10a106','K10a91','K10a24','K10a103','K10a104','K10a118','K10a72','K10a95','K10a66','K10a93','K10a100','K11a3','K11a5','K11a7','K11a9','K11a14','K11a15','K11a17','K11a19','K11a22','K11a24','K11a25','K11a26','K11a28','K11a33','K11a34','K11a35','K11a40','K11a42','K11a44','K11a47','K11a51','K11a53','K11a55','K11a57','K11a62','K11a66','K11a68','K11a71','K11a72','K11a73','K11a74','K11a76','K11a79','K11a80','K11a81','K11a82','K11a83','K11a86','K11a88','K11a92','K11a96','K11a99','K11a106','K11a108','K11a109','K11a112','K11a113']

fibered_knots = [random.choice(fibered) for i in range(10)]

nonfibered_knots = [random.choice(nonfibered) for i in range(10)]

tests = [closed_isosigs,has_internal_homology,links,fibered,nonfibered]

def test(do_tests=[0,1,2,3,4]):

    if 0 in do_tests:
        for name in closed_isosigs:
            print(name)
            T=regina.Triangulation3.fromIsoSig(name)
            W = tnorm.load(T, quiet=True)
            B = W.norm_ball
            assert B.polyhedron().dim() == W.triangulation().homologyH1().rank(), W.manifold().name()+', force_simplicial=False'
            for i in range(W.qtons().size()):
                assert W.euler_char(i) == 2-2*W.genus(i)-W.num_boundary_comps(i), W.manifold().name()+', force_simplicial=False'
            for i in range(len(B.vertices())):
                assert W.is_embedded(B.vertices()[i].qtons_index()) == True, W.manifold().name()+', force_simplicial=False'

            del3 = tnorm.kernel.simplicial.del3_matrix(W.triangulation())
            del2 = tnorm.kernel.simplicial.del2_matrix(W.triangulation())
            assert (del2*del3).is_zero(), W.manifold().name()+', force_simplicial=True'
            for i in range(W.qtons().size()):
                assert (del2*W.simplicial_class(i)).is_zero(), W.manifold().name()+', force_simplicial=True'
            if W.manifold().verify_hyperbolicity()[0]:
                for i in range(W.qtons().size()):
                    assert W.euler_char(i) < 0, W.manifold().name()+', force_simplicial=True'

    
    if 1 in do_tests:
        for name in has_internal_homology:
            print(name)
            W = tnorm.load(name, quiet=True)
            B = W.norm_ball
            assert B.polyhedron().dim() == W.triangulation().homologyH1().rank(), W.manifold().name()+', force_simplicial=False'
            assert W.has_internal_homology(), W.manifold().name()+', force_simplicial=False'
            del3 = tnorm.kernel.simplicial.del3_matrix(W.triangulation())
            del2 = tnorm.kernel.simplicial.del2_matrix(W.triangulation())
            assert (del2*del3).is_zero(), W.manifold().name()+', force_simplicial=False'
            for i in range(W.qtons().size()):
                assert (del2*W.simplicial_class(i)).is_zero(), W.manifold().name()+', force_simplicial=False'
                assert W.euler_char(i) == 2-2*W.genus(i)-W.num_boundary_comps(i), W.manifold().name()+', force_simplicial=False'
            assert B.polyhedron().is_compact(), W.manifold().name()+', force_simplicial=False'
            for i in range(len(B.vertices())):
                assert W.is_embedded(B.vertices()[i].qtons_index()) == True, W.manifold().name()+', force_simplicial=False'
            if W.manifold().verify_hyperbolicity()[0]:
                for i in range(W.qtons().size()):
                    assert W.euler_char(i) < 0, W.manifold().name()+', force_simplicial=False'

            W = tnorm.load(name, quiet=True, force_simplicial_homology=True)
            B = W.norm_ball
            assert B.polyhedron().dim() == W.triangulation().homologyH1().rank(), W.manifold().name()+', force_simplicial=True'
            assert W.has_internal_homology(), W.manifold().name()+', force_simplicial=True'
            del3 = tnorm.kernel.simplicial.del3_matrix(W.triangulation())
            del2 = tnorm.kernel.simplicial.del2_matrix(W.triangulation())
            assert (del2*del3).is_zero(), W.manifold().name()+', force_simplicial=True'
            for i in range(W.qtons().size()):
                assert (del2*W.simplicial_class(i)).is_zero(), W.manifold().name()+', force_simplicial=True'
                assert W.euler_char(i) == 2-2*W.genus(i)-W.num_boundary_comps(i), W.manifold().name()+', force_simplicial=True'
            assert B.polyhedron().is_compact(), W.manifold().name()+', force_simplicial=True'
            if W.manifold().verify_hyperbolicity()[0]:
                for i in range(W.qtons().size()):
                    assert W.euler_char(i) < 0, W.manifold().name()+', force_simplicial=True'

    if 2 in do_tests:
        for name in links:
            print(name)
            W = tnorm.load(name, quiet=True)
            B = W.norm_ball
            assert B.polyhedron().dim() == W.triangulation().homologyH1().rank(), W.manifold().name()+', force_simplicial=False'
            assert not W.has_internal_homology(), W.manifold().name()+', force_simplicial=False'
            for i in range(W.qtons().size()):
                assert W.euler_char(i) == 2-2*W.genus(i)-W.num_boundary_comps(i), W.manifold().name()+', force_simplicial=False'
            assert B.polyhedron().is_compact(), W.manifold().name()+', force_simplicial=False'
            for i in range(len(B.vertices())):
                if B.vertices()[i].qtons_index() != None:
                    assert W.is_embedded(B.vertices()[i].qtons_index()) == True, W.manifold().name()+', force_simplicial=False'
            if W.manifold().verify_hyperbolicity()[0]:
                for i in range(W.qtons().size()):
                    assert W.euler_char(i) < 0, W.manifold().name()+', force_simplicial=False'

            if W.manifold().num_cusps()>1:
                W_sim = tnorm.load(name, quiet=True, force_simplicial_homology=True)
                B_sim = W_sim.norm_ball
                assert B_sim.polyhedron().dim() == W_sim.triangulation().homologyH1().rank(), W_sim.manifold().name()+', force_simplicial=True'
                assert not W_sim.has_internal_homology(), W_sim.manifold().name()+', force_simplicial=True'
                del3 = tnorm.kernel.simplicial.del3_matrix(W_sim.triangulation())
                del2 = tnorm.kernel.simplicial.del2_matrix(W_sim.triangulation())
                assert (del2*del3).is_zero(), W_sim.manifold().name()+', force_simplicial=True'
                for i in range(W_sim.qtons().size()):
                    assert (del2*W_sim.simplicial_class(i)).is_zero(), W_sim.manifold().name()+', force_simplicial=True'
                    assert W_sim.euler_char(i) == 2-2*W_sim.genus(i)-W_sim.num_boundary_comps(i), W_sim.manifold().name()+', force_simplicial=True'
                assert B_sim.polyhedron().is_compact(), W_sim.manifold().name()+', force_simplicial=True'
                P = B.polyhedron()
                if W.manifold().verify_hyperbolicity()[0]:
                    for i in range(W.qtons().size()):
                        assert W.euler_char(i) < 0, W.manifold().name()+', force_simplicial=True'

    if 3 in do_tests:
        for name in fibered_knots:
            print(name)
            W = tnorm.load(name, quiet=True)
            B = W.norm_ball
            assert B.polyhedron().dim() == W.triangulation().homologyH1().rank(), W.manifold().name()+', force_simplicial=False'
            assert not W.has_internal_homology(), W.manifold().name()+', force_simplicial=False'
            for i in range(W.qtons().size()):
                assert W.euler_char(i) == 2-2*W.genus(i)-W.num_boundary_comps(i), W.manifold().name()+', force_simplicial=False'
            assert B.polyhedron().is_compact(), W.manifold().name()+', force_simplicial=False'
            if W.manifold().verify_hyperbolicity()[0]:
                for i in range(W.qtons().size()):
                    assert W.euler_char(i) < 0, W.manifold().name()+', force_simplicial=False'

    if 4 in do_tests:
        for name in nonfibered_knots:
            print(name)
            W = tnorm.load(name, quiet=True)
            B = W.norm_ball
            assert B.polyhedron().dim() == W.triangulation().homologyH1().rank(), W.manifold().name()+', force_simplicial=False'
            assert not W.has_internal_homology(), W.manifold().name()+', force_simplicial=False'
            for i in range(W.qtons().size()):
                assert W.euler_char(i) == 2-2*W.genus(i)-W.num_boundary_comps(i), W.manifold().name()+', force_simplicial=False'
            assert B.polyhedron().is_compact(), W.manifold().name()+', force_simplicial=False'
            for i in range(len(B.vertices())):
                assert W.is_embedded(B.vertices()[i].qtons_index()) == True, W.manifold().name()+', force_simplicial=False'
            if W.manifold().verify_hyperbolicity()[0]:
                for i in range(W.qtons().size()):
                    assert W.euler_char(i) < 0, W.manifold().name()+', force_simplicial=False'        



